import vigorish.database as db
from vigorish.util.result import Result


def update_status_brooks_pitch_logs_for_game_list(scraped_data, db_session, new_brooks_game_ids):
    plog_missing_json, plog_missing_db_records = [], []
    for bb_game_id in new_brooks_game_ids:
        pitch_logs_for_game = scraped_data.get_brooks_pitch_logs_for_game(bb_game_id)
        if not pitch_logs_for_game:
            plog_missing_json.append(bb_game_id)
            continue
        result = update_game_status_records(db_session, pitch_logs_for_game)
        if result.failure:
            plog_missing_db_records.append(bb_game_id)
            continue
        result = create_pitch_appearance_status_records(db_session, pitch_logs_for_game)
        if result.failure:
            return result
        db_session.commit()
    return (
        Result.Ok()
        if not (plog_missing_json or plog_missing_db_records)
        else Result.Fail(_get_error_message(plog_missing_json, plog_missing_db_records))
    )


def _get_error_message(plog_missing_json, plog_missing_db_records):
    errors = []
    if plog_missing_json:
        errors.append(f"Failed to retrieve JSON files for pitch logs: {','.join(plog_missing_json)}")
    if plog_missing_db_records:
        errors.append(f"bb_game_id not found in scrape_status_game table: {','.join(plog_missing_db_records)}")
    return "\n".join(errors)


def update_status_brooks_pitch_logs_for_game(db_session, pitch_logs_for_game):
    result = update_game_status_records(db_session, pitch_logs_for_game)
    if result.failure:
        return result
    result = create_pitch_appearance_status_records(db_session, pitch_logs_for_game)
    if result.failure:
        return result
    db_session.commit()
    return Result.Ok()


def update_game_status_records(db_session, pitch_logs_for_game):
    try:
        bb_game_id = pitch_logs_for_game.bb_game_id
        game_status = db.GameScrapeStatus.find_by_bb_game_id(db_session, bb_game_id)
        if not game_status:
            error = f"scrape_status_game does not contain an entry for bb_game_id: {bb_game_id}"
            return Result.Fail(error)
        game_status.scraped_brooks_pitch_logs = 1
        return Result.Ok()
    except Exception as e:
        return Result.Fail(f"Error: {repr(e)}")


def create_pitch_appearance_status_records(db_session, pitch_logs_for_game):
    season = None
    bbref_game_id = pitch_logs_for_game.bbref_game_id
    all_pitch_app_ids = db.PitchAppScrapeStatus.get_all_pitch_app_ids_for_game(db_session, bbref_game_id)
    scraped_pitch_app_ids = [plog.pitch_app_id for plog in pitch_logs_for_game.pitch_logs]
    update_pitch_app_ids = set(scraped_pitch_app_ids).difference(all_pitch_app_ids)
    for pitch_log in pitch_logs_for_game.pitch_logs:
        if pitch_log.pitch_app_id not in update_pitch_app_ids:
            continue
        if not season:
            season = db.Season.find_by_year(db_session, pitch_log.game_date.year)
        try:
            date_status = db.DateScrapeStatus.find_by_date(db_session, pitch_log.game_date)
            game_status = db.GameScrapeStatus.find_by_bb_game_id(db_session, pitch_log.bb_game_id)
            player_id = db.PlayerId.find_by_mlb_id(db_session, pitch_log.pitcher_id_mlb)
            pitch_app_status = db.PitchAppScrapeStatus()
            pitch_app_status.pitcher_id_mlb = pitch_log.pitcher_id_mlb
            pitch_app_status.pitch_app_id = pitch_log.pitch_app_id
            pitch_app_status.bbref_game_id = pitch_log.bbref_game_id
            pitch_app_status.bb_game_id = pitch_log.bb_game_id
            pitch_app_status.pitch_count_pitch_log = pitch_log.total_pitch_count
            pitch_app_status.scrape_status_game_id = game_status.id
            pitch_app_status.scrape_status_date_id = date_status.id
            pitch_app_status.pitcher_id = player_id.db_player_id
            pitch_app_status.season_id = season.id
            if not pitch_log.parsed_all_info:
                pitch_app_status.scraped_pitchfx = 1
                pitch_app_status.no_pitchfx_data = 1
            db_session.add(pitch_app_status)
        except Exception as e:
            return Result.Fail(f"Error: {repr(e)}")
    return Result.Ok()



import subprocess
from copy import deepcopy

from events import Events
from halo import Halo

from vigorish.cli.components import get_random_cli_color, get_random_dots_spinner
from vigorish.tasks.base import Task
from vigorish.tasks.patch_invalid_pfx import PatchInvalidPitchFxTask
from vigorish.util.result import Result


class PatchAllInvalidPitchFxTask(Task):
    def __init__(self, app):
        super().__init__(app)
        self.patch_invalid_pfx = PatchInvalidPitchFxTask(app)
        self.events = Events(
            (
                "error_occurred",
                "patch_all_invalid_pitchfx_started",
                "patch_all_invalid_pitchfx_complete",
                "create_invalid_pfx_map_start",
                "create_invalid_pfx_map_complete",
                "match_missing_pfx_data_start",
                "match_missing_pfx_data_complete",
                "create_patch_list_start",
                "create_patch_list_complete",
                "combine_scraped_data_start",
                "combine_scraped_data_complete",
            )
        )

    def execute(self, year):
        self.subscribe_to_events()
        all_patch_results = {}
        audit_report = self.scraped_data.get_audit_report()
        self.audit_report_before = deepcopy(audit_report)
        if year not in self.audit_report_before:
            return Result.Fail(f"No games for MLB {year} season have been scraped.")
        game_ids = self.audit_report_before[year].get("invalid_pfx", [])
        if not game_ids:
            return Result.Fail(f"No games for MLB {year} season have invalid pitchfx data.")
        self.events.patch_all_invalid_pitchfx_started()
        self.initialize_spinner(game_ids)
        for num, game_id in enumerate(game_ids, start=1):
            self.spinner.text = self.get_spinner_text(game_id, num, len(game_ids))
            result = self.patch_invalid_pfx.execute(game_id, no_prompts=True)
            if result.failure:
                self.spinner.stop()
                return result
            patch_results = result.value
            all_patch_results[game_id] = patch_results
            self.spinner.text = self.get_spinner_text(game_id, num + 1, len(game_ids))
        self.spinner.stop()
        audit_report = self.scraped_data.get_audit_report()
        self.audit_report_after = deepcopy(audit_report)
        (successful_change, invalid_pfx_change) = self.calculate_games_changed(year)
        self.events.patch_all_invalid_pitchfx_complete()
        self.unsubscribe_from_events()
        return Result.Ok(
            {
                "all_patch_results": all_patch_results,
                "successful_change": successful_change,
                "invalid_pfx_change": invalid_pfx_change,
            }
        )

    def initialize_spinner(self, game_ids):
        subprocess.run(["clear"])
        self.spinner = Halo(spinner=get_random_dots_spinner(), color=get_random_cli_color())
        self.spinner.text = self.get_spinner_text(game_ids[0], 1, len(game_ids))
        self.spinner.start()

    def get_spinner_text(self, game_id, game_count, total_games):
        percent = game_count / float(total_games)
        return f"Attempting to patch invalid PitchFX data (Game ID: {game_id}) {percent:.0%}..."

    def calculate_games_changed(self, year):
        successful_before = self.audit_report_before[year]["successful"]
        successful_after = self.audit_report_after[year]["successful"]
        successful_change = len(successful_after) - len(successful_before)
        invalid_pfx_before = self.audit_report_before[year]["invalid_pfx"]
        invalid_pfx_after = self.audit_report_after[year]["invalid_pfx"]
        invalid_pfx_change = len(invalid_pfx_after) - len(invalid_pfx_before)
        return (successful_change, invalid_pfx_change)

    def handle_error_occurred(self, error_message):
        self.events.error_occurred(error_message)

    def create_invalid_pfx_map_start(self):
        self.events.create_invalid_pfx_map_start()

    def create_invalid_pfx_map_complete(self, invalid_pfx_map):
        self.events.create_invalid_pfx_map_complete(invalid_pfx_map)

    def match_missing_pfx_data_start(self):
        self.events.match_missing_pfx_data_start()

    def match_missing_pfx_data_complete(self, match_dict):
        self.events.match_missing_pfx_data_complete(match_dict)

    def create_patch_list_start(self):
        self.events.create_patch_list_start()

    def create_patch_list_complete(self, patch_list):
        self.events.create_patch_list_complete(patch_list)

    def combine_scraped_data_start(self):
        self.events.combine_scraped_data_start()

    def combine_scraped_data_complete(self, patch_diff_report):
        self.events.combine_scraped_data_complete(patch_diff_report)

    def subscribe_to_events(self):
        self.patch_invalid_pfx.events.error_occurred += self.handle_error_occurred
        self.patch_invalid_pfx.events.create_invalid_pfx_map_start += self.create_invalid_pfx_map_start
        self.patch_invalid_pfx.events.create_invalid_pfx_map_complete += self.create_invalid_pfx_map_complete
        self.patch_invalid_pfx.events.match_missing_pfx_data_start += self.match_missing_pfx_data_start
        self.patch_invalid_pfx.events.match_missing_pfx_data_complete += self.match_missing_pfx_data_complete
        self.patch_invalid_pfx.events.create_patch_list_start += self.create_patch_list_start
        self.patch_invalid_pfx.events.create_patch_list_complete += self.create_patch_list_complete
        self.patch_invalid_pfx.events.combine_scraped_data_start += self.combine_scraped_data_start
        self.patch_invalid_pfx.events.combine_scraped_data_complete += self.combine_scraped_data_complete

    def unsubscribe_from_events(self):
        self.patch_invalid_pfx.events.error_occurred -= self.handle_error_occurred
        self.patch_invalid_pfx.events.create_invalid_pfx_map_start -= self.create_invalid_pfx_map_start
        self.patch_invalid_pfx.events.create_invalid_pfx_map_complete -= self.create_invalid_pfx_map_complete
        self.patch_invalid_pfx.events.match_missing_pfx_data_start -= self.match_missing_pfx_data_start
        self.patch_invalid_pfx.events.match_missing_pfx_data_complete -= self.match_missing_pfx_data_complete
        self.patch_invalid_pfx.events.create_patch_list_start -= self.create_patch_list_start
        self.patch_invalid_pfx.events.create_patch_list_complete -= self.create_patch_list_complete
        self.patch_invalid_pfx.events.combine_scraped_data_start -= self.combine_scraped_data_start
        self.patch_invalid_pfx.events.combine_scraped_data_complete -= self.combine_scraped_data_complete



"""Menu that allows the user to view and modify all settings in vig.config.json."""
from vigorish.cli.menu import Menu
from vigorish.cli.menu_items import ReturnToParent, SetupDatabase
from vigorish.cli.menu_items.admin_tasks import (
    AddToDatabase,
    BackupDatabase,
    CalculatePitchTimes,
    ImportScrapedData,
    NpmInstallUpdate,
    RestoreDatabase,
    SyncScrapedData,
    UpdatePlayerIdMap,
)
from vigorish.constants import EMOJIS
from vigorish.util.result import Result
from vigorish.util.sys_helpers import node_is_installed


class AdminTasksMenu(Menu):
    def __init__(self, app):
        # TODO: Add Admin Task menu item to edit season dates and add new seasons
        super().__init__(app)
        self.menu_text = "Select a task from the list:"
        self.menu_item_text = "Tasks/Admin"
        self.menu_item_emoji = EMOJIS.get("PAGER", "")

    def populate_menu_items(self):
        self.menu_items.clear()
        if node_is_installed():
            self.menu_items.append(NpmInstallUpdate(self.app))
        self.menu_items.append(UpdatePlayerIdMap(self.app))
        self.menu_items.append(SyncScrapedData(self.app))
        if self.app.db_setup_complete:
            self.menu_items.append(ImportScrapedData(self.app))
            self.menu_items.append(AddToDatabase(self.app))
            self.menu_items.append(BackupDatabase(self.app))
            self.menu_items.append(RestoreDatabase(self.app))
            self.menu_items.append(CalculatePitchTimes(self.app))
            self.menu_items.append(SetupDatabase(self.app))
        self.menu_items.append(ReturnToParent(self.app, "Return to Main Menu"))
        self.menu_items.insert(0, ReturnToParent(self.app, "Return to Main Menu"))

    def prompt_user_for_menu_selection(self):
        menu = self._get_bullet_menu()
        menu.pos = self.selected_menu_item_pos
        self.selected_menu_item_text = menu.launch()
        result = self.selected_menu_item.launch()
        exit_menu = self.selected_menu_item.exit_menu
        return Result.Ok(exit_menu) if result.success else result



# flake8: noqa
from vigorish.cli.components.data_set_check import DataSetCheck
from vigorish.cli.components.date_input import DateInput
from vigorish.cli.components.job_name_input import JobNameInput
from vigorish.cli.components.prompts import *
from vigorish.cli.components.util import *



import os
from datetime import datetime
from functools import cached_property
from pathlib import Path
from typing import Optional

from sqlalchemy import create_engine, func
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Session, sessionmaker
from sqlalchemy.schema import Table
from vigorish.data.game_data import GameData

import vigorish.database as db
import vigorish.setup.populate_tables as setup_db
from vigorish.config.config_file import ConfigFile
from vigorish.config.config_setting import ConfigSettingValue, PathConfigSetting
from vigorish.config.dotenv_file import DotEnvFile
from vigorish.config.project_paths import CSV_FOLDER, JSON_FOLDER, SQLITE_DEV_URL, SQLITE_PROD_URL
from vigorish.data.scraped_data import ScrapedData
from vigorish.enums import DataSet
from vigorish.types import AuditReport
from vigorish.util.result import Result


class Vigorish:
    dotenv: DotEnvFile
    config: ConfigFile
    db_engine: Engine
    db_session: Session
    scraped_data: ScrapedData

    def __init__(
        self,
        dotenv_file: Optional[Path] = None,
        db_engine: Optional[Engine] = None,
        db_session: Optional[Session] = None,
    ) -> None:
        self.initialize_app(dotenv_file, db_engine, db_session)
        os.environ["INTERACTIVE_MODE"] = "NO" if "TEST" in os.environ.get("ENV", "DEV") else "YES"

    @property
    def dotenv_filepath(self) -> Path:
        return self.dotenv.dotenv_filepath

    @property
    def db_setup_complete(self) -> bool:
        tables_missing = (
            "player" not in self.db_engine.table_names()
            or "season" not in self.db_engine.table_names()
            or "team" not in self.db_engine.table_names()
        )
        if tables_missing:
            return False
        return (
            self.get_total_number_of_rows(db.Season) > 0
            and self.get_total_number_of_rows(db.Player) > 0
            and self.get_total_number_of_rows(db.Team) > 0
        )

    @cached_property
    def audit_report(self) -> AuditReport:
        return self.scraped_data.get_audit_report()

    def initialize_app(
        self,
        dotenv_file: Optional[Path] = None,
        db_engine: Optional[Engine] = None,
        db_session: Optional[Session] = None,
    ) -> None:
        self.dotenv = DotEnvFile(dotenv_filepath=dotenv_file)
        self.db_url = _get_db_url()
        self.config = ConfigFile()
        self.db_engine = db_engine or self._create_db_engine()
        self.db_session = db_session or self._create_db_session()
        self.scraped_data = ScrapedData(self.db_engine, self.db_session, self.config)

    def get_total_number_of_rows(self, db_table: Table) -> int:
        q = self.db_session.query(db_table)
        count_q = q.statement.with_only_columns([func.count()]).order_by(None)
        return q.session.execute(count_q).scalar()

    def initialize_database(self, csv_folder: Optional[Path] = None, json_folder: Optional[Path] = None) -> Result:
        if not csv_folder:
            csv_folder = CSV_FOLDER
        if not json_folder:
            json_folder = JSON_FOLDER
        self._create_db_schema()
        return setup_db.populate_tables(self, csv_folder, json_folder)

    def prepare_database_for_restore(
        self, csv_folder: Optional[Path] = None, json_folder: Optional[Path] = None
    ) -> Result:
        if not csv_folder:
            csv_folder = CSV_FOLDER
        if not json_folder:
            json_folder = JSON_FOLDER
        self._delete_db_file()
        self.reset_database_connection()
        self._create_db_schema()
        return setup_db.populate_tables_for_restore(self, csv_folder, json_folder)

    def reset_database_connection(self) -> None:
        self.db_session.close()
        self.db_session = None
        self.initialize_app(self.dotenv_filepath)

    def create_scrape_job(
        self, data_set: DataSet, start_date: datetime, end_date: datetime, job_name: Optional[str] = None
    ) -> Result[db.ScrapeJob]:
        result = db.Season.validate_date_range(self.db_session, start_date, end_date)
        if result.failure:
            return result
        season = result.value
        new_job = db.ScrapeJob.from_user_params(self.db_session, data_set, start_date, end_date, season, job_name)
        return Result.Ok(new_job)

    def get_current_setting(self, setting_name, data_set=DataSet.ALL, year=None) -> ConfigSettingValue:
        config_setting = self.config.all_settings.get(setting_name)
        if not config_setting:
            raise ValueError(f"{setting_name} is not a valid configuration setting")
        return (
            config_setting.current_setting(data_set).resolve(year)
            if isinstance(config_setting, PathConfigSetting)
            else config_setting.current_setting(data_set)
        )

    def get_scoreboard_data_for_date(self, game_date: datetime):
        game_ids = db.DateScrapeStatus.get_all_bbref_game_ids_for_date(self.db_session, game_date)
        return [GameData(self, game_id).get_game_data() for game_id in game_ids]

    def _create_db_engine(self) -> Engine:
        return create_engine(self.db_url, connect_args={"check_same_thread": False})

    def _create_db_session(self) -> Session:
        return sessionmaker(bind=self.db_engine)()

    def _create_db_schema(self) -> None:
        db.Base.metadata.drop_all(self.db_engine)
        db.Base.metadata.create_all(self.db_engine)

    def _delete_db_file(self) -> None:
        db_file = Path(self.db_url.replace("sqlite:///", ""))
        if db_file.exists():
            db_file.unlink()


def _get_db_url() -> str:
    db_url = os.getenv("DATABASE_URL", "")
    if db_url and db_url.startswith("/"):
        db_url = f"sqlite:///{db_url}"
    env = os.getenv("ENV", "prod")
    return db_url or (SQLITE_DEV_URL if env == "dev" else SQLITE_PROD_URL)



from dataclasses import dataclass

from dataclass_csv import accept_whitespaces
from sqlalchemy import Column, Float, ForeignKey, Integer, String
from sqlalchemy.orm import relationship

import vigorish.database as db


class BatStats(db.Base):

    __tablename__ = "bat_stats"
    id = Column(Integer, primary_key=True)
    bbref_game_id = Column(String)
    player_id_mlb = Column(Integer)
    player_id_bbref = Column(String)
    player_team_id_bbref = Column(String)
    opponent_team_id_bbref = Column(String)
    is_starter = Column(Integer, default=0)
    bat_order = Column(Integer)
    def_position = Column(String)
    plate_appearances = Column(Integer)
    at_bats = Column(Integer)
    hits = Column(Integer)
    runs_scored = Column(Integer)
    rbis = Column(Integer)
    bases_on_balls = Column(Integer)
    strikeouts = Column(Integer)
    doubles = Column(Integer, default=0)
    triples = Column(Integer, default=0)
    homeruns = Column(Integer, default=0)
    stolen_bases = Column(Integer, default=0)
    caught_stealing = Column(Integer, default=0)
    hit_by_pitch = Column(Integer, default=0)
    intentional_bb = Column(Integer, default=0)
    gdp = Column(Integer, default=0)
    sac_fly = Column(Integer, default=0)
    sac_hit = Column(Integer, default=0)
    avg_to_date = Column(Float)
    obp_to_date = Column(Float)
    slg_to_date = Column(Float)
    ops_to_date = Column(Float)
    total_pitches = Column(Integer)
    total_strikes = Column(Integer)
    wpa_bat = Column(Float)
    avg_lvg_index = Column(Float)
    wpa_bat_pos = Column(Float)
    wpa_bat_neg = Column(Float)
    re24_bat = Column(Float)

    player_id = Column(Integer, ForeignKey("player.id"), index=True)
    player_team_id = Column(Integer, ForeignKey("team.id"))
    opponent_team_id = Column(Integer, ForeignKey("team.id"))
    game_status_id = Column(Integer, ForeignKey("scrape_status_game.id"))
    date_id = Column(Integer, ForeignKey("scrape_status_date.id"))
    season_id = Column(Integer, ForeignKey("season.id"))

    player = relationship("Player", foreign_keys=[player_id], backref="bat_stats")
    player_team = relationship("Team", foreign_keys=[player_team_id], backref="team_bat_stats")
    opponent_team = relationship("Team", foreign_keys=[opponent_team_id], backref="oppp_bat_stats")
    game = relationship("GameScrapeStatus", foreign_keys=[game_status_id])
    date = relationship("DateScrapeStatus", foreign_keys=[date_id])
    season = relationship("Season", foreign_keys=[season_id])

    @classmethod
    def from_dict(cls, bbref_game_id, bat_stats_dict):
        bbref_data = bat_stats_dict.pop("bbref_data", {})
        bbref_data["bbref_game_id"] = bbref_game_id
        bbref_data["player_id_mlb"] = bat_stats_dict["batter_id_mlb"]
        bbref_data["player_id_bbref"] = bat_stats_dict["batter_id_bbref"]
        bbref_data["player_team_id_bbref"] = bat_stats_dict["batter_team_id_bbref"]
        bbref_data["opponent_team_id_bbref"] = bat_stats_dict["opponent_team_id_bbref"]
        bbref_data["is_starter"] = int(bat_stats_dict["is_starter"])
        bbref_data["bat_order"] = bat_stats_dict["bat_order"]
        bbref_data["def_position"] = bat_stats_dict["def_position"]
        for d in bbref_data.pop("details", []):
            if d["stat"] == "2B":
                bbref_data["doubles"] = d["count"]
            if d["stat"] == "3B":
                bbref_data["triples"] = d["count"]
            if d["stat"] == "HR":
                bbref_data["homeruns"] = d["count"]
            if d["stat"] == "SB":
                bbref_data["stolen_bases"] = d["count"]
            if d["stat"] == "CS":
                bbref_data["caught_stealing"] = d["count"]
            if d["stat"] == "HBP":
                bbref_data["hit_by_pitch"] = d["count"]
            if d["stat"] == "IW":
                bbref_data["intentional_bb"] = d["count"]
            if d["stat"] == "GDP":
                bbref_data["gdp"] = d["count"]
            if d["stat"] == "SF":
                bbref_data["sac_fly"] = d["count"]
            if d["stat"] == "SH":
                bbref_data["sac_hit"] = d["count"]
        return cls(**bbref_data)

    @classmethod
    def get_all_seasons_with_data_for_player(cls, db_session, player_id_mlb):
        return {bs.season for bs in db_session.query(cls).filter_by(player_id_mlb=player_id_mlb).all()}


@accept_whitespaces
@dataclass
class BatStatsCsvRow:
    id: int = None
    bbref_game_id: str = None
    player_id_mlb: int = None
    player_id_bbref: str = None
    player_team_id_bbref: str = None
    opponent_team_id_bbref: str = None
    is_starter: int = 0
    bat_order: int = 0
    def_position: str = None
    plate_appearances: int = 0
    at_bats: int = 0
    hits: int = 0
    runs_scored: int = 0
    rbis: int = 0
    bases_on_balls: int = 0
    strikeouts: int = 0
    doubles: int = 0
    triples: int = 0
    homeruns: int = 0
    stolen_bases: int = 0
    caught_stealing: int = 0
    hit_by_pitch: int = 0
    intentional_bb: int = 0
    gdp: int = 0
    sac_fly: int = 0
    sac_hit: int = 0
    avg_to_date: float = 0.0
    obp_to_date: float = 0.0
    slg_to_date: float = 0.0
    ops_to_date: float = 0.0
    total_pitches: int = 0
    total_strikes: int = 0
    wpa_bat: float = 0.0
    avg_lvg_index: float = 0.0
    wpa_bat_pos: float = 0.0
    wpa_bat_neg: float = 0.0
    re24_bat: float = 0.0



"""Scrape brooksbaseball daily dashboard page."""
from datetime import datetime

import w3lib.url
from lxml import html

import vigorish.database as db
from vigorish.constants import BB_BR_TEAM_ID_MAP, PACIFIC_TZ_TEAMS
from vigorish.scrape.brooks_games_for_date.models.game_info import BrooksGameInfo
from vigorish.scrape.brooks_games_for_date.models.games_for_date import BrooksGamesForDate
from vigorish.util.dt_format_strings import DATE_ONLY, DATE_ONLY_TABLE_ID
from vigorish.util.result import Result
from vigorish.util.string_helpers import (
    parse_timestamp,
    validate_bbref_game_id_list,
    validate_brooks_game_id,
)

GAME_TABLE_XPATH = '//td[@class="dashcell"]/table'
GAME_INFO_XPATH = "./tbody//tr[1]//td[1]//text()"
PITCH_LOG_URL_XPATH = './tbody//a[text()="Game Log"]/@href'
KZONE_URL_XPATH = './tbody//a[text()="Strikezone Map"]/@href'


def parse_brooks_dashboard_page(db_session, scraped_html, game_date, url, bbref_games_for_date):
    page_content = html.fromstring(scraped_html, base_url=url)
    games_for_date = BrooksGamesForDate()
    games_for_date.game_date = game_date
    games_for_date.game_date_str = game_date.strftime(DATE_ONLY)
    games_for_date.dashboard_url = url
    games_for_date.game_count = 0
    games = []
    if db.Season.is_this_the_asg_date(db_session, game_date):
        return Result.Ok(games_for_date)
    game_tables = page_content.xpath(GAME_TABLE_XPATH)
    if not game_tables:
        return Result.Ok(games_for_date)
    for i, game in enumerate(game_tables):
        required_game_data = validate_bbref_game_id_list(bbref_games_for_date.all_bbref_game_ids)
        game_info_list = game.xpath(GAME_INFO_XPATH)
        if not game_info_list or len(game_info_list) != 2:
            error = f"Game info table #{i + 1} was not in the expected format"
            return Result.Fail(error)
        game_time = parse_timestamp(game_info_list[1])
        pitchlog_urls = game.xpath(PITCH_LOG_URL_XPATH)
        if not pitchlog_urls:
            result = _no_pitch_logs(game, i + 1, game_date, game_time, required_game_data)
            if result.failure:
                continue
            gameinfo = result.value
            games.append(gameinfo)
            continue
        result = _is_game_required(game_date, game_time, pitchlog_urls[0], required_game_data)
        if result.failure:
            continue
        gameinfo = result.value
        result = _parse_pitch_log_dict(gameinfo, pitchlog_urls)
        if result.failure:
            return result
        gameinfo = result.value
        if all(gameinfo != g for g in games):
            games.append(gameinfo)

    games_for_date.games = games
    games_for_date.game_count = len(games_for_date.games)
    _update_game_ids(games_for_date)
    return Result.Ok(games_for_date)


def _no_pitch_logs(game_table, game_number, game_date, game_time, required_game_data):
    k_zone_urls = game_table.xpath(KZONE_URL_XPATH)
    if not k_zone_urls:
        error = f"Unable to parse bb_game_id for game table #{game_number}"
        return Result.Fail(error)
    result = _parse_gameinfo(game_date, game_time, k_zone_urls[0])
    if result.failure:
        return result
    gameinfo = result.value
    game_is_required = any(gameinfo.home_team_id_bb == req["home_team_id"] for req in required_game_data)
    if not game_is_required:
        return Result.Fail("Game not found in bbref parsed game list.")
    gameinfo.might_be_postponed = True
    gameinfo.pitcher_appearance_count = 0
    return Result.Ok(gameinfo)


def _is_game_required(game_date, game_time, url, required_game_data):
    result = _parse_gameinfo(game_date, game_time, url)
    if result.failure:
        return result
    gameinfo = result.value
    result = validate_brooks_game_id(gameinfo.bb_game_id)
    if result.failure:
        return result
    game_date_check = result.value["game_date"]
    if game_date != game_date_check:
        return Result.Fail("Game did not occur on the current date.")
    game_is_required = any(gameinfo.home_team_id_bb == req["home_team_id"] for req in required_game_data)
    if not game_is_required:
        return Result.Fail("Game not found in bbref parsed game list.")
    gameinfo.might_be_postponed = False
    return Result.Ok(gameinfo)


def _parse_gameinfo(game_date, game_time, url):
    result = _parse_gameinfo_from_url(url)
    if result.failure:
        return result
    gameinfo = result.value
    home_team_id_br = BB_BR_TEAM_ID_MAP.get(gameinfo.home_team_id_bb, gameinfo.home_team_id_bb)
    game_time_hour = (
        game_time["hour"]
        if (game_time["hour"] in [11, 12] or (game_time["hour"] == 10 and home_team_id_br not in PACIFIC_TZ_TEAMS))
        else game_time["hour"] + 12
        if game_time["hour"] in range(1, 11)
        else 0
    )
    gameinfo.game_date_year = game_date.year
    gameinfo.game_date_month = game_date.month
    gameinfo.game_date_day = game_date.day
    gameinfo.game_time_hour = game_time_hour
    gameinfo.game_time_minute = game_time["minute"]
    gameinfo.time_zone_name = "America/New_York"
    return Result.Ok(gameinfo)


def _parse_gameinfo_from_url(url):
    gameid = _parse_gameid_from_url(url)
    if not gameid:
        error = "URL not in expected format, unable to retrieve value of " f'query parameter "game":\n{url}'
        return Result.Fail(error)
    gameinfo = BrooksGameInfo()
    gameinfo.bb_game_id = gameid
    gameinfo.away_team_id_bb = _parse_away_team_from_gameid(gameid)
    gameinfo.home_team_id_bb = _parse_home_team_from_gameid(gameid)
    gameinfo.game_number_this_day = _parse_game_number_from_gameid(gameid)
    return Result.Ok(gameinfo)


def _parse_pitch_log_dict(gameinfo, pitchlog_url_list):
    gameinfo.pitcher_appearance_count = len(pitchlog_url_list)
    gameinfo.pitcher_appearance_dict = {}
    for url in pitchlog_url_list:
        pitcher_id = _parse_pitcherid_from_url(url)
        if pitcher_id is None:
            error = "Unable to parse pitcher_id from url query string:\n" f"url: {url}"
            return Result.Fail(error)
        gameinfo.pitcher_appearance_dict[pitcher_id] = url
    return Result.Ok(gameinfo)


def _parse_gameid_from_url(url):
    gameid = w3lib.url.url_query_parameter(url, "game")
    if not gameid:
        return None
    return gameid[:-1] if gameid.endswith(r"/") else gameid


def _parse_away_team_from_gameid(gameid):
    split = gameid.split("_")
    if len(split) != 7:
        return None
    teama = split[4]
    if len(teama) != 6:
        return None
    return teama[:3].upper()


def _parse_home_team_from_gameid(gameid):
    split = gameid.split("_")
    if len(split) != 7:
        return None
    teamh = split[5]
    if len(teamh) != 6:
        return None
    return teamh[:3].upper()


def _parse_game_number_from_gameid(gameid):
    split = gameid.split("_")
    if len(split) != 7:
        return None
    return int(split[6])


def _parse_pitcherid_from_url(url):
    pitcherid = w3lib.url.url_query_parameter(url, "pitchSel")
    if pitcherid is None:
        return None
    if not pitcherid.endswith(".xml"):
        return None
    return pitcherid[:-4]


def _update_game_ids(games_for_date):
    game_dict = {g.bb_game_id: g for g in games_for_date.games}
    tracker = {g.bb_game_id: False for g in games_for_date.games}
    for bb_gid in sorted(game_dict.keys(), reverse=True):
        if tracker[bb_gid]:
            continue
        g = game_dict[bb_gid]
        game_date = datetime(g.game_date_year, g.game_date_month, g.game_date_day)
        date_str = game_date.strftime(DATE_ONLY_TABLE_ID)

        if g.game_number_this_day == 2:
            g.bbref_game_id = f"{g.home_team_id_bb}{date_str}2"
            game_1_bb_id = f"{bb_gid[:-1]}{1}"
            game_1 = game_dict[game_1_bb_id]
            game_1.bbref_game_id = f"{g.home_team_id_bb}{date_str}1"
            tracker[bb_gid] = True
            tracker[game_1_bb_id] = True
        else:
            g.bbref_game_id = f"{g.home_team_id_bb}{date_str}0"
            tracker[bb_gid] = True



# flake8: noqa
from vigorish.cli.menu_items.admin_tasks.add_to_database import AddToDatabase
from vigorish.cli.menu_items.admin_tasks.backup_database import BackupDatabase
from vigorish.cli.menu_items.admin_tasks.calculate_avg_pitch_times import CalculatePitchTimes
from vigorish.cli.menu_items.admin_tasks.import_scraped_data import ImportScrapedData
from vigorish.cli.menu_items.admin_tasks.npm_install_update import NpmInstallUpdate
from vigorish.cli.menu_items.admin_tasks.restore_database import RestoreDatabase
from vigorish.cli.menu_items.admin_tasks.sync_scraped_data import SyncScrapedData
from vigorish.cli.menu_items.admin_tasks.update_player_id_map import UpdatePlayerIdMap



from abc import ABC, abstractmethod


class Task(ABC):
    """Base class for a task that requires access to the db and application state."""

    def __init__(self, app):
        self.app = app
        self.dotenv = app.dotenv
        self.config = app.config
        self.db_engine = app.db_engine
        self.db_session = app.db_session
        self.scraped_data = app.scraped_data

    @abstractmethod
    def execute(self, *args, **kwargs):
        pass



"""Compiled regular expressions and patterns."""
import re

from vigorish.enums import DataSet, VigFile

TIMESTAMP_REGEX = re.compile(r"(?P<hour>\d?\d):(?P<minute>\d\d)")
JOB_NAME_PATTERN = r"^[\w\s-]+$"
JOB_NAME_REGEX = re.compile(JOB_NAME_PATTERN)
DATE_ONLY_TABLE_ID_REGEX = re.compile(
    r"""
    (?P<year>\d{4,4})
    (?P<month>\d{2,2})
    (?P<day>\d{2,2})
""",
    re.VERBOSE,
)
BBREF_BOXSCORE_URL_PATTERN = r"""
    ^
    https:\/\/www.baseball-reference.com\/boxes\/
    (?P<team_id>[A-Z]{3,3})\/
    (?P<game_id>[0-9A-Z]{12,12}).shtml
    $
"""
BBREF_BOXSCORE_URL_REGEX = re.compile(BBREF_BOXSCORE_URL_PATTERN, re.VERBOSE)

BBREF_DAILY_JSON_FILENAME_REGEX = re.compile(
    r"""
    bbref_games_for_date_
    (?P<year>\d{4,4})-
    (?P<month>\d{2,2})-
    (?P<day>\d{2,2})
""",
    re.VERBOSE,
)
BBREF_DAILY_JSON_FILENAME_REGEX_STRICT = re.compile(
    r"""
    ^
    bbref_games_for_date_
    (?P<year>\d{4,4})-
    (?P<month>\d{2,2})-
    (?P<day>\d{2,2})
    $
""",
    re.VERBOSE,
)
BBREF_DAILY_PATCH_LIST_FILENAME_REGEX = re.compile(
    r"""
    bbref_games_for_date_
    (?P<year>\d{4,4})-
    (?P<month>\d{2,2})-
    (?P<day>\d{2,2})
    _PATCH_LIST
""",
    re.VERBOSE,
)

BB_DAILY_JSON_FILENAME_REGEX = re.compile(
    r"""
    brooks_games_for_date_
    (?P<year>\d{4,4})-
    (?P<month>\d{2,2})-
    (?P<day>\d{2,2})
""",
    re.VERBOSE,
)
BB_DAILY_JSON_FILENAME_REGEX_STRICT = re.compile(
    r"""
    ^
    brooks_games_for_date_
    (?P<year>\d{4,4})-
    (?P<month>\d{2,2})-
    (?P<day>\d{2,2})
    $
""",
    re.VERBOSE,
)
BB_DAILY_PATCH_LIST_FILENAME_REGEX = re.compile(
    r"""
    brooks_games_for_date_
    (?P<year>\d{4,4})-
    (?P<month>\d{2,2})-
    (?P<day>\d{2,2})
    _PATCH_LIST
""",
    re.VERBOSE,
)

PFX_TIMESTAMP_REGEX = re.compile(
    r"""
    (?P<year>\d{2,2})
    (?P<month>\d{2,2})
    (?P<day>\d{2,2})_
    (?P<hour>\d{2,2})
    (?P<minute>\d{2,2})
    (?P<second>\d{2,2})
    (?P<team_id>[a-z]{3,3})
""",
    re.VERBOSE,
)

BBREF_GAME_ID_REGEX = re.compile(
    r"""
    (?P<home_team>[A-Z]{3,3})
    (?P<year>\d{4,4})
    (?P<month>\d{2,2})
    (?P<day>\d{2,2})
    (?P<game_num>\d)
""",
    re.VERBOSE,
)

BBREF_GAME_ID_REGEX_STRICT = re.compile(
    r"""
    ^
    (?P<home_team>[A-Z]{3,3})
    (?P<year>\d{4,4})
    (?P<month>\d{2,2})
    (?P<day>\d{2,2})
    (?P<game_num>\d)
    $
""",
    re.VERBOSE,
)

BBREF_BOXSCORE_PATCH_LIST_FILENAME_REGEX = re.compile(
    r"""
    (?P<home_team>[A-Z]{3,3})
    (?P<year>\d{4,4})
    (?P<month>\d{2,2})
    (?P<day>\d{2,2})
    (?P<game_num>\d)
    _PATCH_LIST
""",
    re.VERBOSE,
)

PITCHFX_LOG_PATCH_LIST_FILENAME_REGEX = BBREF_BOXSCORE_PATCH_LIST_FILENAME_REGEX

COMBINED_DATA_FILENAME_REGEX = re.compile(r"(?P<bbref_game_id>[A-Z]{3,3}\d{9,9})_COMBINED_DATA")

BB_GAME_ID_REGEX = re.compile(
    r"""
    gid_
    (?P<year>\d{4,4})_
    (?P<month>\d{2,2})_
    (?P<day>\d{2,2})_
    (?P<away_team>[a-z]+)mlb_
    (?P<home_team>[a-z]+)mlb_
    (?P<game_num>\d)
""",
    re.VERBOSE,
)

AT_BAT_ID_REGEX = re.compile(
    r"""
    (?P<home_team>[A-Z]{3,3})
    (?P<year>\d{4,4})
    (?P<month>\d{2,2})
    (?P<day>\d{2,2})
    (?P<game_num>\d)_
    (?P<inning>\d{1,2})_
    (?P<pitcher_team>[A-Z]{3,3})_
    (?P<pitcher_mlb_id>\d{6,6})_
    (?P<batter_team>[A-Z]{3,3})_
    (?P<batter_mlb_id>\d{6,6})_
    (?P<at_bat_num>\d)
""",
    re.VERBOSE,
)

PITCH_APP_REGEX = re.compile(
    r"""
    (?P<home_team>[A-Z]{3,3})
    (?P<year>\d{4,4})
    (?P<month>\d{2,2})
    (?P<day>\d{2,2})
    (?P<game_num>\d)
    _(?P<mlb_id>\d{6,6})
""",
    re.VERBOSE,
)

INNING_LABEL_REGEX = re.compile(
    r"""
    (?P<inning_half>TOP|BOT)
    (?P<inning_number>[0-9]{2,2})
""",
    re.VERBOSE,
)

URL_ID_REGEX = {
    VigFile.SCRAPED_HTML: {
        DataSet.BROOKS_GAMES_FOR_DATE: DATE_ONLY_TABLE_ID_REGEX,
        DataSet.BROOKS_PITCH_LOGS: PITCH_APP_REGEX,
        DataSet.BROOKS_PITCHFX: PITCH_APP_REGEX,
        DataSet.BBREF_GAMES_FOR_DATE: DATE_ONLY_TABLE_ID_REGEX,
        DataSet.BBREF_BOXSCORES: BBREF_GAME_ID_REGEX_STRICT,
    },
    VigFile.PARSED_JSON: {
        DataSet.BROOKS_GAMES_FOR_DATE: BB_DAILY_JSON_FILENAME_REGEX_STRICT,
        DataSet.BROOKS_PITCH_LOGS: BB_GAME_ID_REGEX,
        DataSet.BROOKS_PITCHFX: PITCH_APP_REGEX,
        DataSet.BBREF_GAMES_FOR_DATE: BBREF_DAILY_JSON_FILENAME_REGEX_STRICT,
        DataSet.BBREF_BOXSCORES: BBREF_GAME_ID_REGEX_STRICT,
    },
    VigFile.PATCH_LIST: {
        DataSet.BROOKS_GAMES_FOR_DATE: BB_DAILY_PATCH_LIST_FILENAME_REGEX,
        DataSet.BROOKS_PITCHFX: PITCHFX_LOG_PATCH_LIST_FILENAME_REGEX,
        DataSet.BBREF_GAMES_FOR_DATE: BBREF_DAILY_PATCH_LIST_FILENAME_REGEX,
        DataSet.BBREF_BOXSCORES: BBREF_BOXSCORE_PATCH_LIST_FILENAME_REGEX,
    },
    VigFile.COMBINED_GAME_DATA: {DataSet.ALL: COMBINED_DATA_FILENAME_REGEX},
}

URL_ID_CONVERT_REGEX = {
    VigFile.SCRAPED_HTML: {
        DataSet.BROOKS_GAMES_FOR_DATE: DATE_ONLY_TABLE_ID_REGEX,
        DataSet.BBREF_GAMES_FOR_DATE: DATE_ONLY_TABLE_ID_REGEX,
    },
    VigFile.PARSED_JSON: {
        DataSet.BROOKS_GAMES_FOR_DATE: BB_DAILY_JSON_FILENAME_REGEX,
        DataSet.BBREF_GAMES_FOR_DATE: BBREF_DAILY_JSON_FILENAME_REGEX,
    },
    VigFile.COMBINED_GAME_DATA: {DataSet.ALL: COMBINED_DATA_FILENAME_REGEX},
}



"""Menu item that is used to exit the CLI."""
from vigorish.cli.menu_item import MenuItem
from vigorish.constants import EMOJIS
from vigorish.util.result import Result


class ExitProgram(MenuItem):
    def __init__(self, app):
        super().__init__(app)
        self.menu_item_text = "Exit"
        self.menu_item_emoji = EMOJIS.get("EXIT", "")
        self.exit_menu = True

    def launch(self):
        return Result.Ok()



"""Aggregate pitchfx data and play-by-play data into a single object."""
from collections import defaultdict, OrderedDict
from copy import deepcopy
from datetime import datetime, timedelta
from typing import Dict, List

import vigorish.database as db
from vigorish.constants import (
    AT_BAT_RESULTS_ERROR,
    AT_BAT_RESULTS_HBP,
    AT_BAT_RESULTS_HIT,
    AT_BAT_RESULTS_OUT,
    AT_BAT_RESULTS_SAC_FLY,
    AT_BAT_RESULTS_SAC_HIT,
    AT_BAT_RESULTS_STRIKEOUT,
    AT_BAT_RESULTS_UNCLEAR,
    AT_BAT_RESULTS_WALK,
    PLAY_DES_BB_TYPE_FB,
    PLAY_DES_BB_TYPE_GB,
    PLAY_DES_BB_TYPE_LD,
    PLAY_DES_BB_TYPE_PU,
    PPB_PITCH_LOG_DICT,
)
from vigorish.enums import DataSet, PitchType
from vigorish.scrape.bbref_boxscores.models.boxscore import BBRefBoxscore
from vigorish.scrape.brooks_pitchfx.models.pitchfx_log import BrooksPitchFxLog
from vigorish.status.update_status_combined_data import update_pitch_apps_with_combined_data
from vigorish.tasks.base import Task
from vigorish.tasks.scrape_mlb_player_info import ScrapeMlbPlayerInfoTask
from vigorish.util.dt_format_strings import DT_AWARE
from vigorish.util.list_helpers import flatten_list2d
from vigorish.util.result import Result
from vigorish.util.string_helpers import (
    get_brooks_team_id,
    get_inning_id_from_at_bat_id,
    replace_char_with_newlines,
    validate_at_bat_id,
)


class CombineScrapedDataTask(Task):
    bbref_game_id: str
    game_status: db.GameScrapeStatus
    boxscore: BBRefBoxscore
    pitchfx_logs_for_game: BrooksPitchFxLog
    player_id_dict: Dict
    at_bat_ids: List
    at_bat_event_groups: Dict
    all_pfx_data_for_game: List
    removed_pfx: List
    all_removed_pfx: defaultdict
    invalid_pitchfx: defaultdict
    game_events_combined_data: List
    gather_scraped_data_success: bool
    combined_data_success: bool
    save_combined_data_success: bool
    error_messages: List[str]

    def __init__(self, app):
        super().__init__(app)

    @property
    def away_team_id_br(self):
        return self.boxscore.away_team_data.team_id_br

    @property
    def home_team_id_br(self):
        return self.boxscore.home_team_data.team_id_br

    @property
    def game_start_time(self):
        return self.game_status.game_start_time if self.game_status else None

    @property
    def game_start_time_str(self):
        return self.game_start_time.strftime(DT_AWARE) if self.game_status else None

    def execute(self, bbref_game_id, apply_patch_list=True, write_json=True, update_db=True):
        self.bbref_game_id = bbref_game_id
        self.apply_patch_list = apply_patch_list
        self.write_json = write_json
        self.update_db = update_db
        self.error_messages = []
        result = (
            self.gather_scraped_data()
            .on_failure(self.gather_scraped_data_failed)
            .on_success(self.check_pfx_game_start_time)
            .on_success(self.get_all_pbp_events_for_game)
            .on_success(self.get_all_pfx_data_for_game)
            .on_success(self.combine_pbp_events_with_pfx_data)
            .on_success(self.update_boxscore_with_combined_data)
            .on_both(self.update_game_status)
            .on_failure(self.combined_data_failed)
            .on_success(self.save_combined_data)
            .on_failure(self.save_combined_data_failed)
            .on_success(self.check_update_db)
            .on_success(self.update_pitch_app_status)
        )
        return result.value

    def investigate(self, bbref_game_id, apply_patch_list=False):
        self.bbref_game_id = bbref_game_id
        self.apply_patch_list = apply_patch_list
        result = self.gather_scraped_data()
        if result.failure:
            return {"gather_scraped_data_success": False, "error": result.error}
        self.check_pfx_game_start_time()
        audit_results = {
            "gather_scraped_data_success": True,
            "bbref_game_id": self.bbref_game_id,
            "boxscore": self.boxscore,
            "pitchfx_logs_for_game": self.pitchfx_logs_for_game,
        }
        result = self.get_all_pbp_events_for_game()
        if result.failure:
            audit_results["get_all_pbp_events_success"] = False
            audit_results["error_message"] = result.error
            return audit_results
        audit_results["get_all_pbp_events_success"] = True
        audit_results["player_id_dict"] = self.player_id_dict
        audit_results["at_bat_ids"] = self.at_bat_ids
        audit_results["at_bat_event_groups"] = self.at_bat_event_groups
        result = self.get_all_pfx_data_for_game()
        if result.failure:
            audit_results["get_all_pfx_data_success"] = False
            audit_results["error_message"] = result.error
            return audit_results
        audit_results["get_all_pfx_data_success"] = True
        audit_results["all_pfx_data_for_game"] = self.all_pfx_data_for_game
        result = self.combine_pbp_events_with_pfx_data()
        if result.failure:
            audit_results["combine_data_success"] = False
            audit_results["error_message"] = result.error
            return audit_results
        audit_results["combine_data_success"] = True
        audit_results["game_events_combined_data"] = self.game_events_combined_data
        result = self.update_boxscore_with_combined_data()
        if result.failure:
            audit_results["update_boxscore_success"] = False
            audit_results["error_message"] = result.error
            return audit_results
        audit_results["update_boxscore_success"] = True
        audit_results["boxscore"] = result.value
        return audit_results

    def gather_scraped_data(self):
        self.game_status = db.GameScrapeStatus.find_by_bbref_game_id(self.db_session, self.bbref_game_id)
        self.boxscore = self.scraped_data.get_bbref_boxscore(self.bbref_game_id, self.apply_patch_list)
        if not self.boxscore:
            error = f"Failed to retrieve {DataSet.BBREF_BOXSCORES} (URL ID: {self.bbref_game_id})"
            Result.Ok(error)
        result = self.scraped_data.get_all_brooks_pitchfx_logs_for_game(self.bbref_game_id, self.apply_patch_list)
        if result.failure:
            return result
        self.pitchfx_logs_for_game = result.value
        self.avg_pitch_times = db.TimeBetweenPitches.get_latest_results(self.db_session)
        return Result.Ok()

    def check_pfx_game_start_time(self):
        self.gather_scraped_data_success = True
        if not self.game_status.game_start_time:
            self.update_game_start_time()
        for pitchfx_log in self.pitchfx_logs_for_game:
            if not pitchfx_log.game_start_time:
                pitchfx_log.game_date_year = self.game_start_time.year
                pitchfx_log.game_date_month = self.game_start_time.month
                pitchfx_log.game_date_day = self.game_start_time.day
                pitchfx_log.game_time_hour = self.game_start_time.hour
                pitchfx_log.game_time_minute = self.game_start_time.minute
                pitchfx_log.time_zone_name = "America/New_York"
            if all(pfx.game_start_time_str for pfx in pitchfx_log.pitchfx_log):
                continue
            for pfx in pitchfx_log.pitchfx_log:
                if not pfx.game_start_time_str:
                    pfx.game_start_time_str = self.game_start_time_str
        return Result.Ok()

    def update_game_start_time(self):
        all_pfx = [pfx for pfx_log in self.pitchfx_logs_for_game for pfx in pfx_log.pitchfx_log]
        all_pfx.sort(key=lambda x: x.park_sv_id)
        first_pitch_thrown = all_pfx[0].time_pitch_thrown
        game_start_time = (
            first_pitch_thrown.replace(second=0)
            if first_pitch_thrown.second
            else first_pitch_thrown - timedelta(minutes=1)
        )
        self.game_status.game_time_hour = game_start_time.hour
        self.game_status.game_time_minute = game_start_time.minute
        self.game_status.game_time_zone = "America/New_York"
        self.db_session.commit()

    def get_all_pbp_events_for_game(self):
        result = self.get_player_id_dict_for_game()
        if result.failure:
            return result
        self.player_id_dict = result.value
        self.at_bat_ids = []
        self.at_bat_event_groups = {}
        all_events = self.get_all_events()
        at_bat_events = []

        for game_event in all_events:
            if at_bat_events and game_event.inning_label != at_bat_events[-1].inning_label:
                prev_event = at_bat_events[-1]
                self.update_at_bat_event_groups(prev_event, at_bat_events)
                at_bat_events = []
            at_bat_events.append(game_event)
            if self.event_is_player_substitution_or_misc(game_event):
                continue
            if game_event.pitch_sequence:
                game_event.pitch_sequence = game_event.pitch_sequence.replace("^", "")
                result = self.pitch_sequence_is_complete_at_bat(
                    game_event.pitch_sequence, game_event.pbp_table_row_number
                )
                if result.failure:
                    return result
                game_event.is_complete_at_bat = result.value
            else:
                game_event.is_complete_at_bat = False
            if (
                game_event.is_complete_at_bat
                or self.event_resulted_in_third_out(game_event)
                or self.runner_reached_on_interference(game_event)
            ):
                self.add_new_at_bat_event_group(game_event, at_bat_events)
                at_bat_events = []
        if at_bat_events:
            prev_event = at_bat_events[-1]
            self.update_at_bat_event_groups(prev_event, at_bat_events)
        return Result.Ok()

    def get_player_id_dict_for_game(self):
        player_id_dict = {}
        player_name_dict = self.boxscore.player_name_dict
        for name_team_id, bbref_id in player_name_dict.items():
            split = name_team_id.split(",")
            name = split[0].strip()
            team_id = split[1].strip()
            player = db.PlayerId.find_by_bbref_id(self.db_session, bbref_id)
            if not player:
                result = ScrapeMlbPlayerInfoTask(self.app).execute(name, bbref_id, self.boxscore.game_date)
                if result.failure:
                    return result
                player = result.value
            player_id_dict[bbref_id] = {
                "name": name,
                "mlb_id": player.mlb_id,
                "team_id_bbref": team_id,
            }
        return Result.Ok(player_id_dict)

    def get_all_events(self):
        game_events = flatten_list2d([inning.game_events for inning in self.boxscore.innings_list])
        substitutions = flatten_list2d([inning.substitutions for inning in self.boxscore.innings_list])
        misc_events = flatten_list2d([inning.misc_events for inning in self.boxscore.innings_list])
        all_events = game_events + substitutions + misc_events
        all_events.sort(key=lambda x: x.pbp_table_row_number)
        return all_events

    def update_at_bat_event_groups(self, game_event, at_bat_events):
        return (
            self.update_existing_at_bat_event_group(game_event, at_bat_events)
            if self.event_is_player_substitution_or_misc(game_event)
            else self.add_new_at_bat_event_group(game_event, at_bat_events)
        )

    def update_existing_at_bat_event_group(self, game_event, at_bat_events):
        prev_at_bat_id = self.at_bat_ids[-1]
        self.at_bat_event_groups[prev_at_bat_id].append(self.create_bbref_game_event_dict(game_event, prev_at_bat_id))

    def add_new_at_bat_event_group(self, game_event, at_bat_events):
        at_bat_id = self.get_new_at_bat_id(game_event)
        self.at_bat_ids.append(at_bat_id)
        self.at_bat_event_groups[at_bat_id] = [
            self.create_bbref_game_event_dict(event, at_bat_id) for event in at_bat_events
        ]

    def event_is_player_substitution_or_misc(self, event):
        return "BBRefInGameSubstitution" in str(type(event)) or "BBRefPlayByPlayMiscEvent" in str(type(event))

    def event_is_pbp_at_bat(self, event):
        return "BBRefPlayByPlayEvent" in str(type(event))

    def pitch_sequence_is_complete_at_bat(self, pitch_seq, row_num):
        last_pitch = pitch_seq[-1]
        if last_pitch in ["X", "H", "Y"]:
            return Result.Ok(True)
        balls = 0
        strikes = 0
        for pitch in pitch_seq:
            if pitch in ["C", "S", "T", "K", "L", "M", "O", "Q"]:
                strikes += 1
            if pitch in ["F", "R"] and strikes < 2:
                strikes += 1
            if pitch in ["B", "I", "P", "V"]:
                balls += 1
            if pitch in ["U"]:
                error = (
                    f'Error! Unknown pitch type, "{pitch}", occurred in sequence: {pitch_seq} '
                    f"(pbp_table_row# {row_num}, game_id: {self.bbref_game_id})"
                )
                return Result.Fail(error)
        return Result.Ok(True) if strikes == 3 or balls == 4 else Result.Ok(False)

    def event_resulted_in_third_out(self, game_event):
        got_third_out = "O" in game_event.runs_outs_result and game_event.outs_before_play == 2
        got_third_out_dp = "OO" in game_event.runs_outs_result and game_event.outs_before_play == 1
        return got_third_out or got_third_out_dp

    def runner_reached_on_interference(self, game_event):
        return "reached on interference" in game_event.play_description.lower()

    def get_new_at_bat_id(self, game_event):
        instance_num = 0
        at_bat_id = self.get_at_bat_id_for_pbp_event(game_event, instance_num)
        id_exists = at_bat_id in self.at_bat_ids
        while id_exists:
            instance_num += 1
            at_bat_id = self.get_at_bat_id_for_pbp_event(game_event, instance_num)
            id_exists = at_bat_id in self.at_bat_ids
        return at_bat_id

    def create_bbref_game_event_dict(self, event, at_bat_id):
        event_dict = event.as_dict()
        event_dict.pop("__bbref_pbp_game_event__", None)
        event_dict.pop("__bbref_pbp_misc_event__", None)
        event_dict.pop("__bbref_pbp_in_game_substitution__", None)
        event_dict["at_bat_id"] = at_bat_id
        event_dict["event_type"] = event.event_type.name
        if self.event_is_pbp_at_bat(event):
            event_dict["is_complete_at_bat"] = event.is_complete_at_bat
        return event_dict

    def get_at_bat_id_for_pbp_event(self, game_event, instance_number=0):
        inning = f"{int(game_event.inning_label[1:]):02}"
        pteam = get_brooks_team_id(game_event.team_pitching_id_br)
        pid = self.player_id_dict[game_event.pitcher_id_br]["mlb_id"]
        bteam = get_brooks_team_id(game_event.team_batting_id_br)
        bid = self.player_id_dict[game_event.batter_id_br]["mlb_id"]
        return f"{self.bbref_game_id}_{inning}_{pteam}_{pid}_{bteam}_{bid}_{instance_number}"

    def get_all_pfx_data_for_game(self):
        self._assign_at_bat_ids_to_all_pfx_data()
        self._fix_at_bat_ids_with_multiple_pfx_ab_ids()
        self._update_pfx_attributes()
        self._sort_all_pfx_data_by_time_pitch_thrown()
        return Result.Ok()

    def _assign_at_bat_ids_to_all_pfx_data(self):
        all_pfx = [pfx for pfx_log in self.pitchfx_logs_for_game for pfx in pfx_log.pitchfx_log]
        self.all_pfx_data_for_game = list(map(self._set_pfx_at_bat_id, all_pfx))

    def _set_pfx_at_bat_id(self, pfx, instance_number=0):
        inning = f"{pfx.inning:02}"
        pteam = pfx.pitcher_team_id_bb
        pid = pfx.pitcher_id
        bteam = pfx.opponent_team_id_bb
        bid = pfx.batter_id
        pfx.at_bat_id = f"{pfx.bbref_game_id}_{inning}_{pteam}_{pid}_{bteam}_{bid}_{instance_number}"
        return pfx

    def _fix_at_bat_ids_with_multiple_pfx_ab_ids(self):
        all_at_bat_ids = list({pfx.at_bat_id for pfx in self.all_pfx_data_for_game})
        fix_at_bat_ids = filter(self._at_bat_id_contains_multiple_pfx_ab_ids, all_at_bat_ids)
        for at_bat_id in fix_at_bat_ids:
            for i, pfx_bundle in enumerate(self._get_pfx_bundles_for_at_bat_id(at_bat_id)):
                pfx_bundle = list(map(self._set_pfx_at_bat_id, pfx_bundle, [i] * len(pfx_bundle)))

    def _at_bat_id_contains_multiple_pfx_ab_ids(self, at_bat_id):
        return len(self._get_pfx_ab_ids_for_at_bat_id(at_bat_id)) > 1

    def _get_pfx_ab_ids_for_at_bat_id(self, at_bat_id):
        return list({pfx.ab_id for pfx in self._get_pfx_for_at_bat(at_bat_id)})

    def _get_pfx_for_at_bat(self, at_bat_id):
        return [pfx for pfx in self.all_pfx_data_for_game if pfx.at_bat_id == at_bat_id]

    def _get_pfx_bundles_for_at_bat_id(self, at_bat_id):
        return [
            self._get_pfx_for_pfx_ab_id(pfx_ab_id)
            for pfx_ab_id in sorted(self._get_pfx_ab_ids_for_at_bat_id(at_bat_id))
        ]

    def _get_pfx_for_pfx_ab_id(self, pfx_ab_id):
        return [pfx for pfx in self.all_pfx_data_for_game if pfx.ab_id == pfx_ab_id]

    def _update_pfx_attributes(self):
        for pfx in self.all_pfx_data_for_game:
            pfx.inning_id = get_inning_id_from_at_bat_id(pfx.at_bat_id)
            id_dict = db.PlayerId.get_player_ids_from_at_bat_id(self.db_session, pfx.at_bat_id)
            pfx.pitcher_id_bbref = id_dict["pitcher_id_bbref"]
            pfx.batter_id_bbref = id_dict["batter_id_bbref"]
            pfx.db_pitcher_id = id_dict["pitcher_id_db"]
            pfx.db_batter_id = id_dict["batter_id_db"]

    def _sort_all_pfx_data_by_time_pitch_thrown(self):
        all_pfx_data = []
        for pitchfx_log in self.pitchfx_logs_for_game:
            pitchfx_log.at_bat_ids = list({pfx.at_bat_id for pfx in pitchfx_log.pitchfx_log})
            all_pfx_data.extend(pitchfx_log.pitchfx_log)
        self.all_pfx_data_for_game = sorted(all_pfx_data, key=lambda x: (x.ab_id, x.ab_count))

    def combine_pbp_events_with_pfx_data(self):
        (self.at_bat_ids, at_bat_ids_invalid_pfx) = self.reconcile_at_bat_ids()
        all_removed_pfx = {}
        all_pfx_ab_ids = []
        self.game_events_combined_data = []
        for ab_id in self.at_bat_ids:
            error = None
            removed_pfx = []
            all_pbp_events = self.get_all_pbp_events_for_at_bat(ab_id)
            first_pbp_event = all_pbp_events[0]
            final_pbp_event = all_pbp_events[-1]
            runs_outs_result = "".join(event["runs_outs_result"] for event in all_pbp_events)

            result = self.get_total_pitches_in_sequence(final_pbp_event["pitch_sequence"])
            if result.failure:
                return result
            pitch_count = result.value
            pfx_data = self.get_all_pfx_data_for_at_bat(ab_id)
            pfx_ab_id_list = list({pfx["ab_id"] for pfx in pfx_data})
            if not pfx_ab_id_list:
                pfx_ab_id = all_pfx_ab_ids[-1] + 1 if all_pfx_ab_ids else 1
            elif len(pfx_ab_id_list) == 1:
                pfx_ab_id = pfx_ab_id_list[0]
            else:
                return Result.Fail(f"Error! More than one pfx_ab_id found for ab_id: {ab_id}")
            all_pfx_ab_ids.append(pfx_ab_id)
            if pfx_data and pitch_count == 0:
                invalid_ibb_pfx = self.update_invalid_ibb_pfx(pfx_data)
                removed_pfx.extend(invalid_ibb_pfx)
                pfx_data = []
            missing_pitch_numbers = []
            result = self.determine_pfx_sequence(ab_id, pfx_data, pitch_count)
            if result.failure:
                error = result.error
            else:
                (pfx_data, out_of_sequence_pfx, missing_pitch_numbers) = result.value
                removed_pfx.extend(out_of_sequence_pfx)
                if pitch_count > 0 and pfx_data and not missing_pitch_numbers:
                    final_pitch_of_ab = pfx_data[-1]
                    self._update_at_bat_result_stats(final_pitch_of_ab, final_pbp_event)
            (since_start, dur, first_pitch, last_pitch) = self.get_at_bat_duration(pfx_data)
            self.prepare_pfx_data_for_json_serialization(pfx_data)
            if removed_pfx:
                all_removed_pfx[ab_id] = removed_pfx
                self.prepare_pfx_data_for_json_serialization(removed_pfx)
            result = self.describe_at_bat(ab_id, final_pbp_event, pfx_data, missing_pitch_numbers)
            if result.failure:
                return result
            pitch_sequence_description = result.value
            id_dict = db.PlayerId.get_player_ids_from_at_bat_id(self.db_session, ab_id)
            pfx_data = self.determine_bb_type(pfx_data, final_pbp_event["play_description"])
            combined_at_bat_data = {
                "at_bat_id": ab_id,
                "pfx_ab_id": pfx_ab_id,
                "inning_id": get_inning_id_from_at_bat_id(ab_id),
                "pitch_app_id": self.get_pitch_app_id_from_at_bat_id(ab_id),
                "pbp_table_row_number": first_pbp_event["pbp_table_row_number"],
                "pitcher_id_bbref": id_dict["pitcher_id_bbref"],
                "pitcher_id_mlb": id_dict["pitcher_id_mlb"],
                "pitcher_name": id_dict["pitcher_name"],
                "batter_id_bbref": id_dict["batter_id_bbref"],
                "batter_id_mlb": id_dict["batter_id_mlb"],
                "batter_name": id_dict["batter_name"],
                "at_bat_pitchfx_audit": {
                    "pitch_count_bbref": pitch_count,
                    "pitch_count_pitchfx": len(pfx_data),
                    "patched_pitchfx_count": len([pfx for pfx in pfx_data if pfx["is_patched"]]),
                    "missing_pitchfx_count": len(missing_pitch_numbers),
                    "removed_pitchfx_count": len(removed_pfx),
                    "missing_pitch_numbers": missing_pitch_numbers,
                    "pitchfx_error": error is not None,
                    "pitchfx_error_message": error,
                },
                "first_pitch_thrown": first_pitch,
                "last_pitch_thrown": last_pitch,
                "since_game_start": since_start,
                "at_bat_duration": dur,
                "is_complete_at_bat": final_pbp_event["is_complete_at_bat"],
                "score": first_pbp_event["score"],
                "outs_before_play": first_pbp_event["outs_before_play"],
                "runners_on_base": first_pbp_event["runners_on_base"],
                "runs_outs_result": runs_outs_result,
                "play_description": final_pbp_event["play_description"],
                "pitch_sequence_description": pitch_sequence_description,
                "pbp_events": self.at_bat_event_groups[ab_id],
                "pitchfx": pfx_data,
                "removed_pitchfx": removed_pfx,
            }
            self.game_events_combined_data.append(combined_at_bat_data)
        self.save_removed_pfx(all_removed_pfx)
        self.save_invalid_pitchfx(at_bat_ids_invalid_pfx)
        return Result.Ok()

    def reconcile_at_bat_ids(self):
        at_bat_ids_from_box = list(set(self.at_bat_event_groups.keys()))
        at_bat_ids_from_box = self.order_at_bat_ids_by_time(at_bat_ids_from_box)
        at_bat_ids_from_pfx = [pfx.at_bat_id for pfx in self.all_pfx_data_for_game]
        at_bat_ids_invalid_pfx = list(set(at_bat_ids_from_pfx) - set(at_bat_ids_from_box))
        return (at_bat_ids_from_box, at_bat_ids_invalid_pfx)

    def order_at_bat_ids_by_time(self, at_bat_ids):
        game_event_id_map = [
            {
                "at_bat_id": ab_id,
                "pbp_table_row_number": self._get_first_table_row_num_for_at_bat(ab_id),
            }
            for ab_id in at_bat_ids
        ]
        game_event_id_map.sort(key=lambda x: x["pbp_table_row_number"])
        return [id_map["at_bat_id"] for id_map in game_event_id_map]

    def _get_first_table_row_num_for_at_bat(self, at_bat_id):
        return min(game_event["pbp_table_row_number"] for game_event in self.at_bat_event_groups[at_bat_id])

    def get_all_pbp_events_for_at_bat(self, at_bat_id):
        at_bat_events = [event for event in self.at_bat_event_groups[at_bat_id] if event["event_type"] == "AT_BAT"]
        at_bat_events.sort(key=lambda x: x["pbp_table_row_number"])
        return at_bat_events

    def get_all_pfx_data_for_at_bat(self, at_bat_id):
        pfx_for_at_bat = [pfx for pfx in self.all_pfx_data_for_game if pfx.at_bat_id == at_bat_id]
        return self.convert_pfx_list_to_dict_list(pfx_for_at_bat)

    def convert_pfx_list_to_dict_list(self, pfx_list):
        pfx_dict_list = []
        for pfx in pfx_list:
            pfx_dict = pfx.as_dict()
            pfx_dict.pop("__brooks_pitchfx_data__", None)
            pfx_dict["at_bat_id"] = pfx.at_bat_id
            pfx_dict["inning_id"] = pfx.inning_id
            pfx_dict["pitcher_id_bbref"] = pfx.pitcher_id_bbref
            pfx_dict["batter_id_bbref"] = pfx.batter_id_bbref
            pfx_dict["pitcher_id_db"] = pfx.db_pitcher_id
            pfx_dict["batter_id_db"] = pfx.db_batter_id
            pfx_dict["game_start_time"] = pfx.game_start_time
            pfx_dict["time_pitch_thrown"] = pfx.time_pitch_thrown
            pfx_dict["seconds_since_game_start"] = pfx.seconds_since_game_start
            pfx_dict_list.append(pfx_dict)
        pfx_dict_list.sort(key=lambda x: x["ab_count"])
        return pfx_dict_list

    def get_total_pitches_in_sequence(self, pitch_sequence):
        total_pitches = 0
        for abbrev in pitch_sequence:
            result = self.get_pitch_type(abbrev)
            if result.failure:
                return result
            pitch_type = result.value
            total_pitches += pitch_type["pitch_counts"]
        return Result.Ok(total_pitches)

    def get_pitch_type(self, abbrev):
        try:
            pitch_type = PPB_PITCH_LOG_DICT[abbrev]
            return Result.Ok(pitch_type)
        except KeyError as e:
            return Result.Fail(f"Invalid pitch abbreviation: {abbrev}\n{repr(e)}")

    def update_invalid_ibb_pfx(self, invalid_ibb_pfx):
        for pfx in invalid_ibb_pfx:
            pfx["is_invalid_ibb"] = True
        return deepcopy(sorted(invalid_ibb_pfx, key=lambda x: (x["ab_id"], x["ab_count"])))

    def determine_pfx_sequence(self, at_bat_id, pfx_data, pitch_count):
        if pitch_count == 0:
            return Result.Ok(([], [], []))
        prev_ab_id = self.get_prev_at_bat_id(at_bat_id)
        result = self.get_prev_pitch_thrown_time(prev_ab_id)
        if result.failure:
            return result
        prev_pitch_thrown = result.value
        valid_pfx = []
        missing_pitch_numbers = []
        possible_pfx = deepcopy(pfx_data)
        pitch_number_dict = self.get_pitch_number_dict(pfx_data)
        for pitch_num in range(1, pitch_count + 1):
            matches = pitch_number_dict.get(pitch_num, [])
            if not matches:
                missing_pitch_numbers.append(pitch_num)
                continue
            if len(matches) == 1:
                best_pfx = matches[0]
            else:
                result = self.find_best_pfx_for_pitch_number(
                    ab_id=at_bat_id,
                    prev_ab_id=prev_ab_id,
                    pfx_data=matches,
                    pitch_num=pitch_num,
                    prev_pitch_thrown=prev_pitch_thrown,
                )
                if result.failure:
                    return result
                best_pfx = result.value
            valid_pfx.append(best_pfx)
            possible_pfx.remove(best_pfx)
            prev_pitch_thrown = best_pfx["time_pitch_thrown"]
        out_of_sequence_pfx = self.update_out_of_sequence_pfx(possible_pfx)
        return Result.Ok((valid_pfx, out_of_sequence_pfx, missing_pitch_numbers))

    def get_prev_at_bat_id(self, at_bat_id):
        try:
            index = self.at_bat_ids.index(at_bat_id)
            if index == 0:
                return None
            return self.at_bat_ids[index - 1]
        except ValueError:
            return None

    def get_prev_pitch_thrown_time(self, at_bat_id):
        if not at_bat_id:
            return Result.Ok(self.game_start_time)
        result = self.get_game_event(at_bat_id)
        if result.failure:
            return result
        at_bat = result.value
        if not at_bat:
            return Result.Fail(f"No game event found with at bat id {at_bat_id}")
        if not self.pitchfx_data_is_complete(at_bat):
            return Result.Ok(self.game_start_time)
        last_pitch_thrown_str = at_bat["pitchfx"][-1]["time_pitch_thrown_str"]
        if not last_pitch_thrown_str:
            return Result.Ok(self.game_start_time)
        last_pitch_in_at_bat_thrown = datetime.strptime(last_pitch_thrown_str, DT_AWARE)
        return Result.Ok(last_pitch_in_at_bat_thrown)

    def get_pitch_number_dict(self, pfx_data):
        pitch_number_dict = defaultdict(list)
        for pfx in pfx_data:
            pitch_number_dict[pfx["ab_count"]].append(pfx)
        return pitch_number_dict

    def find_best_pfx_for_pitch_number(self, ab_id, prev_ab_id, pfx_data, pitch_num, prev_pitch_thrown):
        pitch_times = self.get_pitch_metrics_prev_at_bat(ab_id, prev_ab_id, pitch_num)
        possible_pfx = []
        for pfx in pfx_data:
            pitch_delta = (pfx["time_pitch_thrown"] - prev_pitch_thrown).total_seconds()
            if pitch_delta < 0 or pitch_delta < int(pitch_times["min"]) or pitch_delta > int(pitch_times["max"]):
                continue
            possible_pfx.append(pfx)
        if not possible_pfx:
            pfx_data.sort(key=lambda x: (-x["has_zone_location"], x["seconds_since_game_start"]))
            return Result.Ok(pfx_data[0])
        if len(possible_pfx) == 1:
            return Result.Ok(possible_pfx[0])
        deltas = [
            {
                "pfx": pfx,
                "has_zone_location": pfx["has_zone_location"],
                "delta": self.delta_avg_time_between_pitches(
                    pitch_times["avg"], prev_pitch_thrown, pfx["time_pitch_thrown"]
                ),
            }
            for pfx in possible_pfx
        ]
        deltas.sort(key=lambda x: (-x["has_zone_location"], x["delta"]))
        return Result.Ok(deltas[0]["pfx"])

    def get_pitch_metrics_prev_at_bat(self, at_bat_id, prev_ab_id, pitch_num):
        if not prev_ab_id:
            return {
                "avg": self.avg_pitch_times["time_between_pitches"]["avg"],
                "min": self.avg_pitch_times["time_between_pitches"]["min"],
                "max": self.avg_pitch_times["time_between_pitches"]["max"],
            }
        same_inning = self.at_bat_ids_are_in_same_inning([at_bat_id, prev_ab_id])
        if pitch_num == 1 and same_inning:
            return {
                "avg": self.avg_pitch_times["time_between_at_bats"]["avg"],
                "min": self.avg_pitch_times["time_between_at_bats"]["min"],
                "max": self.avg_pitch_times["time_between_at_bats"]["max"],
            }
        if pitch_num == 1:
            return {
                "avg": self.avg_pitch_times["time_between_innings"]["avg"],
                "min": self.avg_pitch_times["time_between_innings"]["min"],
                "max": self.avg_pitch_times["time_between_innings"]["max"],
            }
        return {
            "avg": self.avg_pitch_times["time_between_pitches"]["avg"],
            "min": self.avg_pitch_times["time_between_pitches"]["min"],
            "max": self.avg_pitch_times["time_between_pitches"]["max"],
        }

    def delta_avg_time_between_pitches(self, avg, pitch1_thrown, pitch2_thrown):
        return abs(avg - (pitch2_thrown - pitch1_thrown).total_seconds())

    def get_game_event(self, at_bat_id):
        matches = [event for event in self.game_events_combined_data if event["at_bat_id"] == at_bat_id]
        if not matches:
            return Result.Ok(None)
        if len(matches) > 1:
            return Result.Fail(f"Found {len(matches)} at bats with the same id: {at_bat_id}")
        return Result.Ok(matches[0])

    def pitchfx_data_is_complete(self, game_event):
        pitchfx_audit = game_event["at_bat_pitchfx_audit"]
        return (
            pitchfx_audit["pitch_count_pitchfx"] > 0
            and pitchfx_audit["pitch_count_bbref"] == pitchfx_audit["pitch_count_pitchfx"]
            and not pitchfx_audit["pitchfx_error"]
        )

    def at_bat_ids_are_in_same_inning(self, at_bat_ids):
        inning_list = {validate_at_bat_id(ab_id).value["inning_id"] for ab_id in at_bat_ids}
        return len(inning_list) == 1

    def update_out_of_sequence_pfx(self, out_of_sequence_pfx):
        for pfx in out_of_sequence_pfx:
            pfx["is_out_of_sequence"] = True
        return deepcopy(sorted(out_of_sequence_pfx, key=lambda x: (x["ab_id"], x["ab_count"])))

    def _update_at_bat_result_stats(self, final_pitch_of_ab, game_event):
        final_pitch_of_ab["is_final_pitch_of_ab"] = True
        at_bat_result = final_pitch_of_ab["des"].lower()

        if at_bat_result == AT_BAT_RESULTS_UNCLEAR:
            final_pitch_of_ab["ab_result_unclear"] = True
            final_pitch_of_ab["pbp_play_result"] = game_event["play_description"]
            final_pitch_of_ab["pbp_runs_outs_result"] = game_event["pbp_runs_outs_result"]

        if at_bat_result in AT_BAT_RESULTS_OUT:
            final_pitch_of_ab["ab_result_out"] = True

        if at_bat_result in AT_BAT_RESULTS_HIT:
            final_pitch_of_ab["ab_result_hit"] = True
            if at_bat_result == "single":
                final_pitch_of_ab["ab_result_single"] = True
            if at_bat_result == "double":
                final_pitch_of_ab["ab_result_double"] = True
            if at_bat_result == "triple":
                final_pitch_of_ab["ab_result_triple"] = True
            if at_bat_result == "home run":
                final_pitch_of_ab["ab_result_homerun"] = True

        if at_bat_result in AT_BAT_RESULTS_WALK:
            final_pitch_of_ab["ab_result_bb"] = True
            if at_bat_result == "intent walk":
                final_pitch_of_ab["ab_result_ibb"] = True

        if at_bat_result in AT_BAT_RESULTS_STRIKEOUT:
            final_pitch_of_ab["ab_result_k"] = True

        if at_bat_result in AT_BAT_RESULTS_HBP:
            final_pitch_of_ab["ab_result_hbp"] = True

        if at_bat_result in AT_BAT_RESULTS_ERROR:
            final_pitch_of_ab["ab_result_error"] = True

        if at_bat_result in AT_BAT_RESULTS_SAC_HIT:
            final_pitch_of_ab["ab_result_sac_hit"] = True

        if at_bat_result in AT_BAT_RESULTS_SAC_FLY:
            final_pitch_of_ab["ab_result_sac_fly"] = True

    def get_at_bat_duration(self, pfx_data):
        if not pfx_data:
            return (0, 0, None, None)
        since_game_start = pfx_data[0]["seconds_since_game_start"]
        first_pitch_thrown = pfx_data[0]["time_pitch_thrown"]
        last_pitch_thrown = pfx_data[-1]["time_pitch_thrown"]
        if not first_pitch_thrown or not last_pitch_thrown:
            return (since_game_start, 0, None, None)
        at_bat_duration = int((last_pitch_thrown - first_pitch_thrown).total_seconds())
        first_pitch_thrown_str = first_pitch_thrown.strftime(DT_AWARE)
        last_pitch_thrown_str = last_pitch_thrown.strftime(DT_AWARE)
        return (since_game_start, at_bat_duration, first_pitch_thrown_str, last_pitch_thrown_str)

    def get_pitch_app_id_from_at_bat_id(self, at_bat_id):
        return validate_at_bat_id(at_bat_id).value["pitch_app_id"]

    def prepare_pfx_data_for_json_serialization(self, pfx_data):
        for pfx_dict in pfx_data:
            pfx_dict.pop("game_start_time", None)
            pfx_dict.pop("time_pitch_thrown", None)

    def describe_at_bat(self, at_bat_id, final_event_in_at_bat, pfx_data, missing_pitch_numbers):
        if missing_pitch_numbers:
            pfx_data = None
        pitch_sequence = final_event_in_at_bat["pitch_sequence"]
        result = self.get_total_pitches_in_sequence(pitch_sequence)
        if result.failure:
            return result
        total_pitches = result.value
        non_batter_events = self.get_all_other_events_for_at_bat(at_bat_id, final_event_in_at_bat)
        current_pitch = 0
        next_pitch_blocked_by_c = False
        sequence_description = []
        for abbrev in pitch_sequence:
            pitch_number = ""
            pfx_des = ""
            blocked_by_c = ""
            result = self.get_pitch_type(abbrev)
            if result.failure:
                return result
            pitch_type = result.value
            outcome = pitch_type["description"]
            if abbrev == "*":
                next_pitch_blocked_by_c = True
                continue
            if pitch_type["pitch_counts"]:
                current_pitch += 1
                space_count = 1 if total_pitches < 10 or current_pitch >= 10 else 2
                pitch_number = f"Pitch{' '*space_count}{current_pitch}/{total_pitches}"
                if pfx_data:
                    pfx = pfx_data[current_pitch - 1]
                    if abbrev == "X":
                        outcome = pfx["pdes"] if "missing_pdes" not in pfx["pdes"] else pfx["des"]
                    pitch_type = PitchType.from_abbrev(pfx["mlbam_pitch_name"])
                    pfx_des = f'{pfx["start_speed"]:02.0f}mph {pitch_type.print_name}'
                if next_pitch_blocked_by_c:
                    blocked_by_c = "\n(pitch was blocked by catcher)"
                    next_pitch_blocked_by_c = False
                sequence_description.append((pitch_number, f"{outcome}{blocked_by_c}", pfx_des))
                continue
            if abbrev == ".":
                outcome = self.get_next_event_description(non_batter_events, outcome)
            sequence_description.append(("", outcome, ""))
        while any(not event_dict["processed"] for event_dict in non_batter_events.values()):
            outcome = self.get_next_event_description(non_batter_events)
            if outcome:
                sequence_description.append(("", outcome, ""))
        outcome = replace_char_with_newlines(final_event_in_at_bat["play_description"], ";")
        sequence_description.append(("Result", outcome, ""))
        return Result.Ok(sequence_description)

    def get_all_other_events_for_at_bat(self, at_bat_id, final_event_this_at_bat):
        all_other_events = list(self.at_bat_event_groups[at_bat_id])
        all_other_events.sort(key=lambda x: x["pbp_table_row_number"])
        all_other_events.remove(final_event_this_at_bat)
        if not all_other_events:
            return {}
        non_batter_events = OrderedDict()
        for num, event in enumerate(all_other_events, start=1):
            non_batter_events[num] = {
                "processed": False,
                "event": event,
            }
        return non_batter_events

    def get_next_event_description(self, non_batter_events, default_outcome=""):
        outcome = default_outcome
        for event_dict in non_batter_events.values():
            if not event_dict["processed"]:
                event = event_dict["event"]
                outcome = (
                    f'({event["play_description"]})'
                    if event["event_type"] == "AT_BAT"
                    else f'({event["sub_description"]})'
                    if event["event_type"] == "SUBSTITUTION"
                    else f'({event["description"]})'
                )
                event_dict["processed"] = True
                break
        return outcome.strip(".")

    def determine_bb_type(self, pfx_data, play_description):
        pfx = [pfx for pfx in pfx_data if pfx["type"] == "X"]
        if not pfx:
            return pfx_data
        pfx = pfx[0]
        pfx["is_batted_ball"] = True
        if any(bb_type in play_description for bb_type in PLAY_DES_BB_TYPE_FB):
            pfx["is_fly_ball"] = True
        if any(bb_type in play_description for bb_type in PLAY_DES_BB_TYPE_GB):
            pfx["is_ground_ball"] = True
        if any(bb_type in play_description for bb_type in PLAY_DES_BB_TYPE_LD):
            pfx["is_line_drive"] = True
        if any(bb_type in play_description for bb_type in PLAY_DES_BB_TYPE_PU):
            pfx["is_pop_up"] = True
        return pfx_data

    def save_invalid_pitchfx(self, at_bat_ids_invalid_pfx):
        self.invalid_pitchfx = defaultdict(dict)
        for ab_id in self.order_at_bat_ids_by_park_sv_id(at_bat_ids_invalid_pfx):
            inning_id = get_inning_id_from_at_bat_id(ab_id)
            id_dict = db.PlayerId.get_player_ids_from_at_bat_id(self.db_session, ab_id)
            pfx_data = self.get_all_pfx_data_for_at_bat(ab_id)
            pfx_ab_id = 0
            pfx_ab_id_list = list({pfx["ab_id"] for pfx in pfx_data})
            if pfx_ab_id_list and len(pfx_ab_id_list) == 1:
                pfx_ab_id = pfx_ab_id_list[0]
            pitch_count = max(pfx["ab_total"] for pfx in pfx_data)
            out_of_sequence_pfx = []
            missing_pitch_numbers = []
            result = self.determine_pfx_sequence(ab_id, pfx_data, pitch_count)
            if result.success:
                (pfx_data, out_of_sequence_pfx, missing_pitch_numbers) = result.value
            self.prepare_pfx_data_for_json_serialization(pfx_data)
            if len(out_of_sequence_pfx) > 0:
                self.prepare_pfx_data_for_json_serialization(out_of_sequence_pfx)
            at_bat_data = {
                "at_bat_id": ab_id,
                "pfx_ab_id": pfx_ab_id,
                "inning_id": inning_id,
                "pitch_app_id": self.get_pitch_app_id_from_at_bat_id(ab_id),
                "pitcher_id_bbref": id_dict["pitcher_id_bbref"],
                "pitcher_id_mlb": id_dict["pitcher_id_mlb"],
                "pitcher_name": id_dict["pitcher_name"],
                "batter_id_bbref": id_dict["batter_id_bbref"],
                "batter_id_mlb": id_dict["batter_id_mlb"],
                "batter_name": id_dict["batter_name"],
                "at_bat_pitchfx_audit": {
                    "pitch_count_bbref": 0,
                    "pitch_count_pitchfx": len(pfx_data),
                    "patched_pitchfx_count": len([pfx for pfx in pfx_data if pfx["is_patched"]]),
                    "missing_pitchfx_count": len(missing_pitch_numbers),
                    "removed_pitchfx_count": len(out_of_sequence_pfx),
                    "missing_pitch_numbers": missing_pitch_numbers,
                    "pitchfx_error": result.error if result.failure else None,
                    "pitchfx_error_message": result.error,
                },
                "pitchfx": pfx_data,
                "removed_pitchfx": out_of_sequence_pfx,
            }
            self.invalid_pitchfx[inning_id][ab_id] = at_bat_data

    def save_removed_pfx(self, removed_pfx_dict):
        self.all_removed_pfx = defaultdict(dict)
        for ab_id, removed_pfx in removed_pfx_dict.items():
            inning_id = get_inning_id_from_at_bat_id(ab_id)
            id_dict = db.PlayerId.get_player_ids_from_at_bat_id(self.db_session, ab_id)
            self.prepare_pfx_data_for_json_serialization(removed_pfx)
            at_bat_data = {
                "at_bat_id": ab_id,
                "inning_id": inning_id,
                "pitch_app_id": self.get_pitch_app_id_from_at_bat_id(ab_id),
                "pitcher_id_bbref": id_dict["pitcher_id_bbref"],
                "pitcher_id_mlb": id_dict["pitcher_id_mlb"],
                "pitcher_name": id_dict["pitcher_name"],
                "batter_id_bbref": id_dict["batter_id_bbref"],
                "batter_id_mlb": id_dict["batter_id_mlb"],
                "batter_name": id_dict["batter_name"],
                "pitchfx": removed_pfx,
            }
            self.all_removed_pfx[inning_id][ab_id] = at_bat_data

    def update_boxscore_with_combined_data(self):
        updated_innings_list = [self.update_inning_with_combined_data(inning) for inning in self.boxscore.innings_list]
        (pitch_stats_away, pitch_stats_home) = self.update_all_pitch_stats()
        (bat_stats_away, bat_stats_home) = self.update_all_bat_stats()

        away_team_data = self.boxscore.away_team_data.as_dict()
        away_team_data.pop("__bbref_boxscore_team_data__", None)
        away_team_data.pop("batting_stats", None)
        away_team_data.pop("pitching_stats", None)
        away_team_data["batting_stats"] = bat_stats_away
        away_team_data["pitching_stats"] = pitch_stats_away

        home_team_data = self.boxscore.home_team_data.as_dict()
        home_team_data.pop("__bbref_boxscore_team_data__", None)
        home_team_data.pop("batting_stats", None)
        home_team_data.pop("pitching_stats", None)
        home_team_data["batting_stats"] = bat_stats_home
        home_team_data["pitching_stats"] = pitch_stats_home

        game_meta_info = self.boxscore.game_meta_info.as_dict()
        game_meta_info.pop("__bbref_boxscore_meta__", None)
        game_meta_info["game_time_hour"] = self.game_start_time.hour
        game_meta_info["game_time_minute"] = self.game_start_time.minute
        game_meta_info["game_date_time_str"] = self.game_start_time.strftime(DT_AWARE)
        game_meta_info["umpires"] = self.boxscore.as_dict()["umpires"]

        pitchfx_vs_bbref_audit = self.audit_pitchfx_vs_bbref_data(pitch_stats_away, pitch_stats_home)

        updated_boxscore_dict = {
            "bbref_game_id": self.bbref_game_id,
            "boxscore_url": self.boxscore.boxscore_url,
            "pitchfx_vs_bbref_audit": pitchfx_vs_bbref_audit,
            "game_meta_info": game_meta_info,
            "away_team_data": away_team_data,
            "home_team_data": home_team_data,
            "play_by_play_data": updated_innings_list,
            "removed_pitchfx": self.all_removed_pfx,
            "invalid_pitchfx": self.invalid_pitchfx,
            "player_id_dict": self.player_id_dict,
        }
        return Result.Ok(updated_boxscore_dict)

    def update_inning_with_combined_data(self, inning):
        inning_events = [event for event in self.game_events_combined_data if event["inning_id"] == inning.inning_id]
        inning_totals = {
            "inning_total_runs": inning.inning_total_runs,
            "inning_total_hits": inning.inning_total_hits,
            "inning_total_errors": inning.inning_total_errors,
            "inning_total_left_on_base": inning.inning_total_left_on_base,
            "away_team_runs_after_inning": inning.away_team_runs_after_inning,
            "home_team_runs_after_inning": inning.home_team_runs_after_inning,
        }
        return {
            "inning_id": inning.inning_id,
            "inning_label": inning.inning_label,
            "begin_inning_summary": inning.begin_inning_summary,
            "end_inning_summary": inning.end_inning_summary,
            "inning_totals": inning_totals,
            "inning_pitchfx_audit": self.generate_audit_report_for_events(inning_events),
            "inning_events": inning_events,
        }

    def generate_audit_report_for_events(self, game_events):
        pitch_count_bbref = sum(event["at_bat_pitchfx_audit"]["pitch_count_bbref"] for event in game_events)
        pitch_count_pitchfx = sum(event["at_bat_pitchfx_audit"]["pitch_count_pitchfx"] for event in game_events)
        patched_pitchfx_count = sum(event["at_bat_pitchfx_audit"]["patched_pitchfx_count"] for event in game_events)
        missing_pitchfx_count = sum(event["at_bat_pitchfx_audit"]["missing_pitchfx_count"] for event in game_events)
        removed_pitchfx_count = sum(event["at_bat_pitchfx_audit"]["removed_pitchfx_count"] for event in game_events)
        pitchfx_error = any(event["at_bat_pitchfx_audit"]["pitchfx_error"] for event in game_events)

        at_bat_ids_pitchfx_complete = list(
            {
                event["at_bat_id"]
                for event in game_events
                if event["is_complete_at_bat"]
                and event["at_bat_pitchfx_audit"]["missing_pitchfx_count"] == 0
                and not event["at_bat_pitchfx_audit"]["pitchfx_error"]
            }
        )
        at_bat_ids_patched_pitchfx = list(
            {event["at_bat_id"] for event in game_events if event["at_bat_pitchfx_audit"]["patched_pitchfx_count"] > 0}
        )
        at_bat_ids_missing_pitchfx = list(
            {event["at_bat_id"] for event in game_events if event["at_bat_pitchfx_audit"]["missing_pitchfx_count"] > 0}
        )
        at_bat_ids_removed_pitchfx = list(
            {event["at_bat_id"] for event in game_events if event["at_bat_pitchfx_audit"]["removed_pitchfx_count"] > 0}
        )
        at_bat_ids_pitchfx_error = list(
            {event["at_bat_id"] for event in game_events if event["at_bat_pitchfx_audit"]["pitchfx_error"]}
        )

        at_bat_ids_pitchfx_complete = self.order_at_bat_ids_by_time(at_bat_ids_pitchfx_complete)
        at_bat_ids_patched_pitchfx = self.order_at_bat_ids_by_time(at_bat_ids_patched_pitchfx)
        at_bat_ids_missing_pitchfx = self.order_at_bat_ids_by_time(at_bat_ids_missing_pitchfx)
        at_bat_ids_removed_pitchfx = self.order_at_bat_ids_by_time(at_bat_ids_removed_pitchfx)
        at_bat_ids_pitchfx_error = self.order_at_bat_ids_by_time(at_bat_ids_pitchfx_error)

        return {
            "pitchfx_error": pitchfx_error,
            "pitch_count_bbref": pitch_count_bbref,
            "pitch_count_pitchfx": pitch_count_pitchfx,
            "patched_pitchfx_count": patched_pitchfx_count,
            "missing_pitchfx_count": missing_pitchfx_count,
            "removed_pitchfx_count": removed_pitchfx_count,
            "total_at_bats_pitchfx_complete": len(at_bat_ids_pitchfx_complete),
            "total_at_bats_patched_pitchfx": len(at_bat_ids_patched_pitchfx),
            "total_at_bats_missing_pitchfx": len(at_bat_ids_missing_pitchfx),
            "total_at_bats_removed_pitchfx": len(at_bat_ids_removed_pitchfx),
            "total_at_bats_pitchfx_error": len(at_bat_ids_pitchfx_error),
            "at_bat_ids_pitchfx_complete": at_bat_ids_pitchfx_complete,
            "at_bat_ids_patched_pitchfx": at_bat_ids_patched_pitchfx,
            "at_bat_ids_missing_pitchfx": at_bat_ids_missing_pitchfx,
            "at_bat_ids_removed_pitchfx": at_bat_ids_removed_pitchfx,
            "at_bat_ids_pitchfx_error": at_bat_ids_pitchfx_error,
        }

    def update_all_pitch_stats(self):
        pitch_stats_dict = {}
        all_bbref_pitch_stats = deepcopy(self.boxscore.away_team_data.pitching_stats)
        all_bbref_pitch_stats.extend(deepcopy(self.boxscore.home_team_data.pitching_stats))
        for pitch_stats in all_bbref_pitch_stats:
            mlb_id = self.player_id_dict[pitch_stats.player_id_br]["mlb_id"]
            pitch_stats_dict[mlb_id] = pitch_stats
        updated_pitcher_ids = []
        updated_pitching_stats = defaultdict(list)
        for pfx_log in self.pitchfx_logs_for_game:
            pitch_stats = pitch_stats_dict.pop(pfx_log.pitcher_id_mlb, None)
            if not pitch_stats:
                error = f"Error retrieving boxscore stats for pitch app: {pfx_log.pitch_app_id}"
                return Result.Fail(error)
            (team_id, updated_stats) = self.update_player_pitch_stats(pfx_log, pitch_stats)
            updated_pitching_stats[team_id].append(updated_stats)
            updated_pitcher_ids.append(pfx_log.pitcher_id_mlb)
        for pitch_stats in pitch_stats_dict.values():
            (team_id, updated_stats) = self.update_player_pitch_stats_no_pfx(pitch_stats)
            updated_pitching_stats[team_id].append(updated_stats)
            pitcher_id_mlb = self.player_id_dict[pitch_stats.player_id_br].get("mlb_id")
            updated_pitcher_ids.append(pitcher_id_mlb)
        pitcher_ids_invalid_pfx = self.get_pitcher_ids_with_invalid_pfx()
        invalid_pitcher_ids = list(set(pitcher_ids_invalid_pfx) - set(updated_pitcher_ids))
        if invalid_pitcher_ids:
            raise NotImplementedError(
                "The code for this condition was removed, create a test case for "
                f"{self.bbref_game_id}, using these pitcher_ids {invalid_pitcher_ids}."
            )
        return (
            updated_pitching_stats[self.away_team_id_br],
            updated_pitching_stats[self.home_team_id_br],
        )

    def update_player_pitch_stats(self, pfx_log, pitch_stats):
        bbref_data = pitch_stats.as_dict()
        bbref_data.pop("player_id_br", None)
        bbref_data.pop("player_team_id_br", None)
        bbref_data.pop("opponent_team_id_br", None)
        pitcher_events = [
            game_event
            for game_event in self.game_events_combined_data
            if game_event["pitcher_id_mlb"] == pfx_log.pitcher_id_mlb
        ]
        batters_faced_pfx = len([event for event in pitcher_events if event["is_complete_at_bat"]])
        audit_report = self.generate_audit_report_for_events(pitcher_events)
        pitch_count_by_inning = self.get_pitch_count_by_inning(pitcher_events)
        pitch_count_by_pitch_type = self.get_pitch_count_by_pitch_type(pitcher_events)
        invalid_pfx = self.get_invalid_pfx_data_for_pitcher(pfx_log.pitcher_id_mlb)
        pitch_app_pitchfx_audit = {
            "invalid_pitchfx": invalid_pfx["invalid_pitchfx"],
            "pitchfx_error": audit_report["pitchfx_error"],
            "pitch_count_bbref": audit_report["pitch_count_bbref"],
            "pitch_count_pitchfx": audit_report["pitch_count_pitchfx"],
            "batters_faced_bbref": pitch_stats.batters_faced,
            "batters_faced_pitchfx": batters_faced_pfx,
            "patched_pitchfx_count": audit_report["patched_pitchfx_count"],
            "missing_pitchfx_count": audit_report["missing_pitchfx_count"],
            "removed_pitchfx_count": audit_report["removed_pitchfx_count"],
            "invalid_pitchfx_count": invalid_pfx["invalid_pitchfx_count"],
            "total_at_bats_pitchfx_complete": audit_report["total_at_bats_pitchfx_complete"],
            "total_at_bats_patched_pitchfx": audit_report["total_at_bats_patched_pitchfx"],
            "total_at_bats_missing_pitchfx": audit_report["total_at_bats_missing_pitchfx"],
            "total_at_bats_removed_pitchfx": audit_report["total_at_bats_removed_pitchfx"],
            "total_at_bats_pitchfx_error": audit_report["total_at_bats_pitchfx_error"],
            "total_at_bats_invalid_pitchfx": invalid_pfx["total_at_bats_invalid_pitchfx"],
            "at_bat_ids_pitchfx_complete": audit_report["at_bat_ids_pitchfx_complete"],
            "at_bat_ids_patched_pitchfx": audit_report["at_bat_ids_patched_pitchfx"],
            "at_bat_ids_missing_pitchfx": audit_report["at_bat_ids_missing_pitchfx"],
            "at_bat_ids_removed_pitchfx": audit_report["at_bat_ids_removed_pitchfx"],
            "at_bat_ids_pitchfx_error": audit_report["at_bat_ids_pitchfx_error"],
            "at_bat_ids_invalid_pitchfx": invalid_pfx["at_bat_ids_invalid_pitchfx"],
        }

        updated_stats = {
            "pitcher_name": pfx_log.pitcher_name,
            "pitcher_id_mlb": pfx_log.pitcher_id_mlb,
            "pitcher_id_bbref": pitch_stats.player_id_br,
            "pitch_app_id": pfx_log.pitch_app_id,
            "pitcher_team_id_bb": pfx_log.pitcher_team_id_bb,
            "pitcher_team_id_bbref": pitch_stats.player_team_id_br,
            "opponent_team_id_bb": pfx_log.opponent_team_id_bb,
            "opponent_team_id_bbref": pitch_stats.opponent_team_id_br,
            "bb_game_id": pfx_log.bb_game_id,
            "bbref_game_id": pfx_log.bbref_game_id,
            "pitch_count_by_inning": pitch_count_by_inning,
            "pitch_count_by_pitch_type": pitch_count_by_pitch_type,
            "pitch_app_pitchfx_audit": pitch_app_pitchfx_audit,
            "bbref_data": bbref_data,
        }
        return (pitch_stats.player_team_id_br, updated_stats)

    def get_pitch_count_by_inning(self, pitcher_events):
        all_pfx = flatten_list2d([event["pitchfx"] for event in pitcher_events])
        unordered = defaultdict(int)
        for pfx in all_pfx:
            unordered[pfx["inning"]] += 1
        pitch_count_by_inning = OrderedDict()
        for k in sorted(unordered.keys()):
            pitch_count_by_inning[k] = unordered[k]
        return pitch_count_by_inning

    def get_pitch_count_by_pitch_type(self, pitcher_events):
        all_pfx = flatten_list2d([event["pitchfx"] for event in pitcher_events])
        pitch_count_unordered = defaultdict(int)
        for pfx in all_pfx:
            pitch_count_unordered[pfx["mlbam_pitch_name"]] += 1
        pitch_count_ordered = OrderedDict()
        ptype_tuples = [(pitch_type, count) for pitch_type, count in pitch_count_unordered.items()]
        for t in sorted(ptype_tuples, key=lambda x: x[1], reverse=True):
            pitch_count_ordered[t[0]] = t[1]
        return pitch_count_ordered

    def get_invalid_pfx_data_for_pitcher(self, pitcher_id_mlb):
        invalid_pfx = {
            "invalid_pitchfx": False,
            "invalid_pitchfx_count": 0,
            "total_at_bats_invalid_pitchfx": 0,
            "at_bat_ids_invalid_pitchfx": [],
        }
        if not self.invalid_pitchfx:
            return invalid_pfx
        pfx_data = flatten_list2d(
            [
                at_bat_data["pitchfx"]
                for invalid_pfx_at_bat_dict in self.invalid_pitchfx.values()
                for at_bat_data in invalid_pfx_at_bat_dict.values()
                if at_bat_data["pitcher_id_mlb"] == pitcher_id_mlb
            ]
        )
        if not pfx_data:
            return invalid_pfx
        at_bat_ids_invalid_pfx = list({pfx["at_bat_id"] for pfx in pfx_data})
        at_bat_ids_invalid_pfx = self.order_at_bat_ids_by_park_sv_id(at_bat_ids_invalid_pfx)
        return {
            "invalid_pitchfx": True,
            "invalid_pitchfx_count": len(pfx_data),
            "total_at_bats_invalid_pitchfx": len(at_bat_ids_invalid_pfx),
            "at_bat_ids_invalid_pitchfx": at_bat_ids_invalid_pfx,
        }

    def get_pitcher_ids_with_invalid_pfx(self):
        pitcher_ids = {
            at_bat_data["pitcher_id_mlb"]
            for invalid_pfx_at_bat_dict in self.invalid_pitchfx.values()
            for at_bat_data in invalid_pfx_at_bat_dict.values()
        }
        return list(pitcher_ids)

    def order_at_bat_ids_by_park_sv_id(self, at_bat_ids):
        park_sv_id_map = [
            {
                "at_bat_id": ab_id,
                "park_sv_id": min(pfx.park_sv_id for pfx in self.all_pfx_data_for_game if pfx.at_bat_id == ab_id),
            }
            for ab_id in at_bat_ids
        ]
        park_sv_id_map.sort(key=lambda x: x["park_sv_id"])
        return [id_map["at_bat_id"] for id_map in park_sv_id_map]

    def update_player_pitch_stats_no_pfx(self, pitch_stats):
        bbref_id = pitch_stats.player_id_br
        mlb_id = self.player_id_dict[bbref_id].get("mlb_id", "")
        pitcher_events = [
            game_event for game_event in self.game_events_combined_data if game_event["pitcher_id_mlb"] == mlb_id
        ]
        at_bat_ids_missing_pitchfx = list({event["at_bat_id"] for event in pitcher_events})
        at_bat_ids_missing_pitchfx = self.order_at_bat_ids_by_time(at_bat_ids_missing_pitchfx)
        bbref_data = pitch_stats.as_dict()
        bbref_data.pop("player_id_br", None)
        bbref_data.pop("player_team_id_br", None)
        bbref_data.pop("opponent_team_id_br", None)

        pitch_app_pitchfx_audit = {
            "invalid_pitchfx": False,
            "pitchfx_error": False,
            "pitch_count_bbref": pitch_stats.pitch_count,
            "pitch_count_pitchfx": 0,
            "batters_faced_bbref": pitch_stats.batters_faced,
            "batters_faced_pitchfx": 0,
            "patched_pitchfx_count": 0,
            "missing_pitchfx_count": pitch_stats.pitch_count,
            "removed_pitchfx_count": 0,
            "invalid_pitchfx_count": 0,
            "total_at_bats_pitchfx_complete": 0,
            "total_at_bats_patched_pitchfx": 0,
            "total_at_bats_missing_pitchfx": len(at_bat_ids_missing_pitchfx),
            "total_at_bats_removed_pitchfx": 0,
            "total_at_bats_pitchfx_error": 0,
            "total_at_bats_invalid_pitchfx": 0,
            "at_bat_ids_pitchfx_complete": [],
            "at_bat_ids_patched_pitchfx": [],
            "at_bat_ids_missing_pitchfx": at_bat_ids_missing_pitchfx,
            "at_bat_ids_removed_pitchfx": [],
            "at_bat_ids_pitchfx_error": [],
            "at_bat_ids_invalid_pitchfx": [],
        }

        updated_stats = {
            "pitcher_name": self.player_id_dict[bbref_id].get("name", ""),
            "pitcher_id_mlb": mlb_id,
            "pitcher_id_bbref": bbref_id,
            "pitch_app_id": f"{self.bbref_game_id}_{mlb_id}",
            "pitcher_team_id_bb": get_brooks_team_id(pitch_stats.player_team_id_br),
            "pitcher_team_id_bbref": pitch_stats.player_team_id_br,
            "opponent_team_id_bb": get_brooks_team_id(pitch_stats.opponent_team_id_br),
            "opponent_team_id_bbref": pitch_stats.opponent_team_id_br,
            "bb_game_id": self.boxscore.bb_game_id,
            "bbref_game_id": self.bbref_game_id,
            "pitch_count_by_inning": [],
            "pitch_count_by_pitch_type": [],
            "pitch_app_pitchfx_audit": pitch_app_pitchfx_audit,
            "bbref_data": bbref_data,
        }
        return (pitch_stats.player_team_id_br, updated_stats)

    def update_all_bat_stats(self):
        all_bbref_bat_stats = self.boxscore.away_team_data.batting_stats
        all_bbref_bat_stats.extend(self.boxscore.home_team_data.batting_stats)
        updated_batting_stats = defaultdict(list)
        for bat_stats in all_bbref_bat_stats:
            bbref_id = bat_stats.player_id_br
            mlb_id = self.player_id_dict[bbref_id]["mlb_id"]
            (team_id, updated_stats) = self.update_player_bat_stats(bbref_id, mlb_id, bat_stats)
            updated_batting_stats[team_id].append(updated_stats)
        return (
            updated_batting_stats[self.away_team_id_br],
            updated_batting_stats[self.home_team_id_br],
        )

    def update_player_bat_stats(self, bbref_id, mlb_id, bat_stats):
        bbref_data = bat_stats.as_dict()
        bbref_data.pop("player_id_br", None)
        batter_team_id_bbref = bbref_data.pop("player_team_id_br", None)
        opponent_team_id_bbref = bbref_data.pop("opponent_team_id_br", None)
        batter_events = [
            game_event for game_event in self.game_events_combined_data if game_event["batter_id_mlb"] == mlb_id
        ]
        all_at_bat_ids = [event["at_bat_id"] for event in batter_events]
        incomplete_at_bat_ids = [event["at_bat_id"] for event in batter_events if not event["is_complete_at_bat"]]
        updated_stats = {
            "batter_name": self.player_id_dict[bbref_id]["name"],
            "batter_id_mlb": mlb_id,
            "batter_id_bbref": bbref_id,
            "batter_team_id_bb": get_brooks_team_id(batter_team_id_bbref),
            "batter_team_id_bbref": batter_team_id_bbref,
            "opponent_team_id_bb": get_brooks_team_id(batter_team_id_bbref),
            "opponent_team_id_bbref": opponent_team_id_bbref,
            "total_pbp_events": len(all_at_bat_ids),
            "total_incomplete_at_bats": len(incomplete_at_bat_ids),
            "total_plate_appearances": bbref_data["plate_appearances"],
            "at_bat_ids": all_at_bat_ids,
            "incomplete_at_bat_ids": incomplete_at_bat_ids,
            "bbref_data": bbref_data,
        }
        return (batter_team_id_bbref, updated_stats)

    def audit_pitchfx_vs_bbref_data(self, away_team_pitching_stats, home_team_pitching_stats):
        batters_faced_bbref_home = sum(
            pitch_stats["pitch_app_pitchfx_audit"]["batters_faced_bbref"] for pitch_stats in home_team_pitching_stats
        )
        batters_faced_bbref_away = sum(
            pitch_stats["pitch_app_pitchfx_audit"]["batters_faced_bbref"] for pitch_stats in away_team_pitching_stats
        )
        batters_faced_bbref = batters_faced_bbref_home + batters_faced_bbref_away

        batters_faced_pitchfx_home = sum(
            pitch_stats["pitch_app_pitchfx_audit"]["batters_faced_pitchfx"] for pitch_stats in home_team_pitching_stats
        )
        batters_faced_pitchfx_away = sum(
            pitch_stats["pitch_app_pitchfx_audit"]["batters_faced_pitchfx"] for pitch_stats in away_team_pitching_stats
        )
        batters_faced_pitchfx = batters_faced_pitchfx_home + batters_faced_pitchfx_away

        pitch_count_bbref_stats_table_home = sum(
            pitch_stats["bbref_data"]["pitch_count"] for pitch_stats in home_team_pitching_stats
        )
        pitch_count_bbref_stats_table_away = sum(
            pitch_stats["bbref_data"]["pitch_count"] for pitch_stats in away_team_pitching_stats
        )
        pitch_count_bbref_stats_table = pitch_count_bbref_stats_table_home + pitch_count_bbref_stats_table_away

        audit_report = self.generate_audit_report_for_events(self.game_events_combined_data)

        at_bat_ids_invalid_pfx = [
            at_bat_ids
            for invalid_pfx_at_bat_dict in self.invalid_pitchfx.values()
            for at_bat_ids in invalid_pfx_at_bat_dict.keys()
        ]
        at_bat_ids_invalid_pfx = self.order_at_bat_ids_by_park_sv_id(at_bat_ids_invalid_pfx)
        total_pitches_invalid_pfx = sum(
            len(at_bat_data["pitchfx"])
            for invalid_pfx_at_bat_dict in self.invalid_pitchfx.values()
            for at_bat_data in invalid_pfx_at_bat_dict.values()
        )
        invalid_pfx = {
            "invalid_pitchfx": bool(self.invalid_pitchfx),
            "invalid_pitchfx_count": total_pitches_invalid_pfx,
            "total_at_bats_invalid_pitchfx": len(at_bat_ids_invalid_pfx),
            "at_bat_ids_invalid_pitchfx": at_bat_ids_invalid_pfx,
        }

        return {
            "invalid_pitchfx": invalid_pfx["invalid_pitchfx"],
            "pitchfx_error": audit_report["pitchfx_error"],
            "pitch_count_bbref_stats_table": pitch_count_bbref_stats_table,
            "pitch_count_bbref": audit_report["pitch_count_bbref"],
            "pitch_count_pitchfx": audit_report["pitch_count_pitchfx"],
            "batters_faced_bbref": batters_faced_bbref,
            "batters_faced_pitchfx": batters_faced_pitchfx,
            "patched_pitchfx_count": audit_report["patched_pitchfx_count"],
            "missing_pitchfx_count": audit_report["missing_pitchfx_count"],
            "removed_pitchfx_count": audit_report["removed_pitchfx_count"],
            "invalid_pitchfx_count": invalid_pfx["invalid_pitchfx_count"],
            "total_at_bats_pitchfx_complete": audit_report["total_at_bats_pitchfx_complete"],
            "total_at_bats_patched_pitchfx": audit_report["total_at_bats_patched_pitchfx"],
            "total_at_bats_missing_pitchfx": audit_report["total_at_bats_missing_pitchfx"],
            "total_at_bats_removed_pitchfx": audit_report["total_at_bats_removed_pitchfx"],
            "total_at_bats_pitchfx_error": audit_report["total_at_bats_pitchfx_error"],
            "total_at_bats_invalid_pitchfx": invalid_pfx["total_at_bats_invalid_pitchfx"],
            "at_bat_ids_pitchfx_complete": audit_report["at_bat_ids_pitchfx_complete"],
            "at_bat_ids_patched_pitchfx": audit_report["at_bat_ids_patched_pitchfx"],
            "at_bat_ids_missing_pitchfx": audit_report["at_bat_ids_missing_pitchfx"],
            "at_bat_ids_removed_pitchfx": audit_report["at_bat_ids_removed_pitchfx"],
            "at_bat_ids_pitchfx_error": audit_report["at_bat_ids_pitchfx_error"],
            "at_bat_ids_invalid_pitchfx": invalid_pfx["at_bat_ids_invalid_pitchfx"],
        }

    def gather_scraped_data_failed(self, error):
        self.gather_scraped_data_success = False
        self.update_db = False
        self.write_json = False
        self.error_messages.append(error)
        result = Result.Fail("")
        result.value = {
            "gather_scraped_data_success": self.gather_scraped_data_success,
            "error": "\n".join(self.error_messages),
        }
        return result

    def combined_data_failed(self, error):
        self.combined_data_success = False
        self.write_json = False
        self.error_messages.append(error)
        result = Result.Fail("")
        result.value = {
            "gather_scraped_data_success": self.gather_scraped_data_success,
            "combined_data_success": self.combined_data_success,
            "error": "\n".join(self.error_messages),
        }
        return result

    def update_game_status(self, result):
        if self.update_db:
            self.game_status.combined_data_success = 1 if result.success else 0
            self.game_status.combined_data_fail = 0 if result.success else 1
            self.db_session.commit()
        return result

    def save_combined_data(self, combined_data):
        self.combined_data_success = True
        self.combined_data = combined_data
        if self.write_json:
            return self.scraped_data.save_combined_game_data(combined_data)
        return Result.Ok()

    def save_combined_data_failed(self, error):
        self.save_combined_data_success = False
        self.error_messages.append(error)
        result = Result.Fail("")
        result.value = {
            "gather_scraped_data_success": self.gather_scraped_data_success,
            "combined_data_success": self.combined_data_success,
            "save_combined_data_success": self.save_combined_data_success,
            "error": "\n".join(self.error_messages),
        }
        return result

    def check_update_db(self, value):
        self.save_combined_data_success = True
        if self.update_db:
            return Result.Ok()
        result = Result.Fail("")
        result.value = {
            "gather_scraped_data_success": self.gather_scraped_data_success,
            "combined_data_success": self.combined_data_success,
            "save_combined_data_success": self.save_combined_data_success,
            "boxscore": self.combined_data,
        }
        return result

    def update_pitch_app_status(self):
        result = update_pitch_apps_with_combined_data(self.db_session, self.combined_data)
        if result.failure:
            self.error_messages.append(result.error)
            result.value = {
                "gather_scraped_data_success": self.gather_scraped_data_success,
                "combined_data_success": self.combined_data_success,
                "save_combined_data_success": self.save_combined_data_success,
                "boxscore": self.combined_data,
                "update_pitch_apps_success": False,
                "error": "\n".join(self.error_messages),
            }
            return result
        results = {
            "gather_scraped_data_success": self.gather_scraped_data_success,
            "combined_data_success": self.combined_data_success,
            "save_combined_data_success": self.save_combined_data_success,
            "boxscore": self.combined_data,
            "update_pitch_apps_success": True,
            "results": result.value,
        }
        result.value = results
        return result



"""Retrieve a response from URL, will retry up to 5 times if the request does not succeed."""
from requests import get as requests_get
from requests.exceptions import ConnectionError, HTTPError, RequestException, Timeout

from vigorish.util.decorators import retry, RetryLimitExceededError
from vigorish.util.result import Result


def request_url_with_retries(url):
    @retry(max_attempts=5, delay=5, exceptions=(ConnectionError, HTTPError, Timeout, RequestException))
    def request_url(url):
        response = requests_get(url)
        response.raise_for_status()
        return response

    try:
        response = request_url(url)
        return Result.Ok(response)
    except RetryLimitExceededError as e:
        return Result.Fail(repr(e))



"""A single pitching appearance scraped from brooksbaseball.com."""
from dataclasses import dataclass, field
from datetime import datetime
from typing import Dict

from dateutil import tz


@dataclass
class BrooksPitchLog:
    """A single pitching appearance scraped from brooksbaseball.com."""

    parsed_all_info: bool = False
    pitcher_name: str = ""
    pitcher_id_mlb: int = 0
    pitch_app_id: str = ""
    total_pitch_count: int = 0
    pitch_count_by_inning: Dict[str, int] = field(default_factory=dict)
    pitcher_team_id_bb: str = ""
    opponent_team_id_bb: str = ""
    bb_game_id: str = ""
    bbref_game_id: str = ""
    game_date_year: str = ""
    game_date_month: str = ""
    game_date_day: str = ""
    game_time_hour: str = ""
    game_time_minute: str = ""
    time_zone_name: str = ""
    pitchfx_url: str = ""
    pitch_log_url: str = ""

    @property
    def game_date(self):
        return datetime(self.game_date_year, self.game_date_month, self.game_date_day).date()

    @property
    def game_start_time(self):
        game_start = datetime(
            year=self.game_date.year,
            month=self.game_date.month,
            day=self.game_date.day,
            hour=self.game_time_hour,
            minute=self.game_time_minute,
        )
        return (
            game_start.replace(tzinfo=tz.gettz(self.time_zone_name))
            if self.game_time_hour != 0 or self.game_time_minute != 0
            else None
        )

    def as_dict(self):
        """Convert pitch log to a dictionary."""
        return {
            "__brooks_pitch_log__": True,
            "parsed_all_info": self.parsed_all_info,
            "pitcher_name": f"{self.pitcher_name}",
            "pitcher_id_mlb": self.pitcher_id_mlb,
            "pitch_app_id": f"{self.pitch_app_id}",
            "total_pitch_count": self.total_pitch_count,
            "pitch_count_by_inning": self.pitch_count_by_inning,
            "pitcher_team_id_bb": f"{self.pitcher_team_id_bb}",
            "opponent_team_id_bb": f"{self.opponent_team_id_bb}",
            "bb_game_id": f"{self.bb_game_id}",
            "bbref_game_id": f"{self.bbref_game_id}",
            "game_date_year": int(self.game_date_year),
            "game_date_month": int(self.game_date_month),
            "game_date_day": int(self.game_date_day),
            "game_time_hour": int(self.game_time_hour),
            "game_time_minute": int(self.game_time_minute),
            "time_zone_name": self.time_zone_name,
            "pitchfx_url": f"{self.pitchfx_url}",
            "pitch_log_url": f"{self.pitch_log_url}",
        }



import json
from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass, field
from typing import List

from vigorish.enums import DataSet
from vigorish.util.list_helpers import as_dict_list
from vigorish.util.result import Result


@dataclass
class Patch(ABC):
    patch_id: str = field(repr=False, init=False)
    data_set: DataSet = field(repr=False, init=False)

    def as_dict(self):
        patch_dict = {self.patch_id: True}
        patch_dict.update(asdict(self))
        patch_dict.pop("patch_id")
        patch_dict.pop("data_set")
        return patch_dict

    def as_json(self):
        return json.dumps(self.as_dict(), indent=2)

    @abstractmethod
    def apply(self, data):
        pass


@dataclass
class PatchList(ABC):
    patch_list_id: str = field(repr=False, init=False)
    data_set: DataSet = field(repr=False, init=False)
    url_id: str
    patch_list: List[Patch] = field(repr=False, default_factory=list)

    def as_dict(self):
        return {
            self.patch_list_id: True,
            "url_id": self.url_id,
            "patch_list": as_dict_list(self.patch_list),
        }

    def as_json(self):
        return json.dumps(self.as_dict(), indent=2)

    def apply(self, data):
        for patch in self.patch_list:
            result = patch.apply(data)
            if result.failure:
                return result
            data = result.value
        return Result.Ok(data)



"""Utility functions that interact with the terminal."""
import os
import platform
import re
import subprocess
from datetime import datetime, timezone
from pathlib import Path
from typing import Tuple, Union
from zipfile import ZipFile

from Naked.toolshed.shell import execute as execute_shell_command
from Naked.toolshed.shell import execute_js

from vigorish.util.datetime_util import dtaware_fromtimestamp
from vigorish.util.dt_format_strings import DT_NAIVE
from vigorish.util.result import Result

ONE_KB = 1000
ONE_MB = ONE_KB * 1000
ONE_GB = ONE_MB * 1000


def run_command(command, cwd=None, shell=True, text=True):  # pragma: no cover
    process = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        cwd=cwd,
        shell=shell,
        text=text,
    )
    for line in iter(process.stdout.readline, ""):
        if line:
            line = line[:-1] if line.endswith("\n") else line
            yield line
    print()


def execute_nodejs_script(script_file_path, script_args):  # pragma: no cover
    result = validate_file_path(script_file_path)
    if result.failure:
        return result
    valid_filepath = result.value
    if program_is_installed("node"):
        success = execute_js(str(valid_filepath), arguments=script_args)
    elif program_is_installed("nodejs"):
        success = execute_shell_command(f"nodejs {valid_filepath} {script_args}")
    else:
        return Result.Fail("Node.js is NOT installed!")
    return Result.Ok() if success else Result.Fail("nodejs script failed")


def node_is_installed():  # pragma: no cover
    return any(program_is_installed(node_alias) for node_alias in ["node", "nodejs"])


def program_is_installed(exe_name, version_option="--version"):  # pragma: no cover
    try:
        check_version = subprocess.run(
            [exe_name, version_option],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            encoding="utf-8",
        )
        output = check_version.stdout.strip()
        return bool(re.compile(r"[vV]?\d+\.").search(output))
    except FileNotFoundError:
        return False


def node_modules_folder_exists():  # pragma: no cover
    app_folder = Path(__file__).parent.parent
    node_modules_folder = app_folder.joinpath("nightmarejs/node_modules")
    return node_modules_folder.exists()


def is_windows():  # pragma: no cover
    return any(platform.win32_ver())


def get_file_size_bytes(filepath):  # pragma: no cover
    result = validate_file_path(filepath)
    if result.failure:
        return result
    valid_filepath = result.value
    return Result.Ok(valid_filepath.stat().st_size)


def get_last_mod_time_utc(filepath):
    return dtaware_fromtimestamp(filepath.stat().st_mtime, use_tz=timezone.utc)


def validate_folder_path(input_path: Union[Path, str]):
    if not input_path:
        return Result.Fail("NoneType or empty string is not a valid folder path.")
    if isinstance(input_path, str):
        folderpath = Path(input_path)
    elif not isinstance(input_path, Path):
        error = f'"input_path" parameter must be str or Path value (not "{type(input_path)}").'
        return Result.Fail(error)
    else:
        folderpath = input_path
    if not folderpath.exists():
        return Result.Fail(f'Directory does NOT exist: "{folderpath}"')
    if not folderpath.is_dir():
        return Result.Fail(f'The provided path is NOT a directory: "{folderpath}"')
    if is_windows() and folderpath.is_reserved():
        return Result.Fail(f'The provided path is reserved under Windows: "{folderpath}"')
    return Result.Ok(folderpath)


def validate_file_path(input_path: Union[Path, str]):
    if not input_path:
        return Result.Fail("NoneType or empty string is not a valid file path.")
    if isinstance(input_path, str):
        filepath = Path(input_path)
    elif not isinstance(input_path, Path):
        error = f'"input_path" parameter must be str or Path value (not "{type(input_path)}").'
        return Result.Fail(error)
    else:
        filepath = input_path
    if not filepath.exists():
        return Result.Fail(f'File does not exist: "{filepath}"')
    if not filepath.is_file():
        return Result.Fail(f'The provided path is NOT a file: "{filepath}"')
    return Result.Ok(filepath)


def get_terminal_size(fallback: Tuple[int, int] = (80, 24)):  # pragma: no cover
    columns, rows = fallback
    for i in range(3):
        try:
            columns, rows = os.get_terminal_size(i)
        except OSError:
            continue
        break
    return columns, rows


def get_terminal_width():  # pragma: no cover
    columns, _ = get_terminal_size()
    return columns


def zip_file_report(zip_file_path):
    report = []
    with ZipFile(zip_file_path, "r") as zip:
        zip_original_size = 0
        for info in zip.infolist():
            zip_original_size += info.file_size
            compression_ratio = 1 - info.compress_size / float(info.file_size)
            (compressed_str, original_str) = align_file_sizes(info.compress_size, info.file_size)
            file_info = (
                f"\tFilename.......: {Path(info.filename).name}\n"
                f"\tModified.......: {datetime(*info.date_time).strftime(DT_NAIVE)} UTC\n"
                f"\tZipped Size....: {compressed_str} ({compression_ratio:.0%} Reduction)\n"
                f"\tOriginal Size..: {original_str}\n"
            )
            report.append(file_info)
        zip_compressed_size = zip_file_path.stat().st_size
        compression_ratio = 1 - zip_compressed_size / float(zip_original_size)
        (compressed_str, original_str) = align_file_sizes(zip_compressed_size, zip_original_size)
        zip_mod_timestamp = zip_file_path.stat().st_mtime
        zip_file_modified = dtaware_fromtimestamp(zip_mod_timestamp, use_tz=timezone.utc)
        file_info = (
            f"Filename.......: {Path(zip_file_path).name}\n"
            f"Modified.......: {zip_file_modified.strftime(DT_NAIVE)} UTC\n"
            f"Zipped Size....: {compressed_str} ({compression_ratio:.0%} Reduction)\n"
            f"Original Size..: {original_str}\n"
        )
        report.insert(0, f"##### ZIP FILE REPORT: {zip_file_path.name} #####\n")
        report.insert(1, file_info)
        report.insert(2, "\t##### ZIP FILE CONTENTS #####\n")
    return "\n".join(report)


def align_file_sizes(compressed_size, original_size):
    compressed_str = file_size_str(compressed_size)
    original_str = file_size_str(original_size)
    if len(original_str) == len(compressed_str):
        return (compressed_str, original_str)
    if len(original_str) > len(compressed_str):
        pad_length = len(original_str) - len(compressed_str)
        compressed_str = f"{' '*pad_length}{compressed_str}"
    else:
        pad_length = len(compressed_str) - len(original_str)
        original_str = f"{' '*pad_length}{original_str}"
    return (compressed_str, original_str)


def file_size_str(size_bytes):
    return (
        f"{size_bytes/float(ONE_GB):.1f} GB"
        if size_bytes > ONE_GB
        else f"{size_bytes/float(ONE_MB):.1f} MB"
        if size_bytes > ONE_MB
        else f"{size_bytes/float(ONE_KB):.1f} KB"
        if size_bytes > ONE_KB
        else f"{size_bytes} bytes"
    )



"""Populate team table with initial data."""
from dataclasses import dataclass

from dataclass_csv import accept_whitespaces, DataclassReader
from tqdm import tqdm

import vigorish.database as db
from vigorish.util.result import Result


@accept_whitespaces
@dataclass
class TeamCsvRow:
    yearID: int = None
    lgID: str = None
    teamID: str = None
    franchID: str = None
    divID: str = None
    Rank: int = None
    G: int = None
    Ghome: int = None
    W: int = None
    L: int = None
    DivWin: str = None
    WCWin: str = None
    LgWin: str = None
    WSWin: str = None
    R: int = None
    AB: int = None
    H: int = None
    doubles: int = None
    triples: int = None
    HR: int = None
    BB: int = None
    SO: int = None
    SB: int = None
    CS: int = None
    HBP: int = None
    SF: int = None
    RA: int = None
    ER: int = None
    ERA: float = None
    CG: int = None
    SHO: int = None
    SV: int = None
    IPouts: int = None
    HA: int = None
    HRA: int = None
    BBA: int = None
    SOA: int = None
    E: int = None
    DP: int = None
    FP: float = None
    name: str = None
    park: str = None
    attendance: int = None
    BPF: int = None
    PPF: int = None
    teamIDBR: str = None
    teamIDlahman45: str = None
    teamIDretro: str = None


TEAMS_CSV_FILENAME = "Teams.csv"


def populate_teams(db_session, csv_folder):
    """Populate team table with initial data."""
    result = import_teams_csv(db_session, csv_folder)
    if result.failure:
        return result
    db_session.commit()
    return Result.Ok()


def import_teams_csv(db_session, csv_folder):
    team_csv_file = csv_folder.joinpath(TEAMS_CSV_FILENAME)
    csv_text = team_csv_file.read_text()
    total_rows = len([row for row in csv_text.split("\n") if row])
    try:
        with open(team_csv_file) as team_csv:
            reader = DataclassReader(team_csv, TeamCsvRow)
            reader.map("2B").to("doubles")
            reader.map("3B").to("triples")
            with tqdm(
                total=total_rows,
                desc="Populating team table..........",
                unit="row",
                mininterval=0.12,
                maxinterval=5,
                unit_scale=True,
                ncols=90,
            ) as pbar:
                for row in reader:
                    t = db.Team(
                        year=row.yearID,
                        league=row.lgID,
                        team_id=row.teamID,
                        franch_id=row.franchID,
                        division=row.divID,
                        games=row.G,
                        games_at_home=row.Ghome,
                        wins=row.W,
                        losses=row.L,
                        runs=row.R,
                        at_bats=row.AB,
                        hits=row.H,
                        doubles=row.doubles,
                        triples=row.triples,
                        homeruns=row.HR,
                        base_on_balls=row.BB,
                        strikeouts=row.SO,
                        stolen_bases=row.SB,
                        caught_stealing=row.CS,
                        runs_against=row.RA,
                        earned_runs=row.ER,
                        saves=row.SV,
                        ip_outs=row.IPouts,
                        errors=row.E,
                        name=row.name,
                        park=row.park,
                        batting_park_factor=row.BPF,
                        pitching_park_factor=row.PPF,
                        team_id_br=row.teamIDBR,
                        team_id_retro=row.teamIDretro,
                    )
                    db_session.add(t)
                    pbar.update()
        return Result.Ok()
    except Exception as e:
        error = f"Error: {repr(e)}"
        db_session.rollback()
        return Result.Fail(error)



"""Base task that defines the template method for a scrape task."""
from abc import ABC, abstractmethod
from functools import partial
from signal import SIGINT, signal
from sys import exit

from getch import pause
from halo import Halo

from vigorish.cli.components import print_message
from vigorish.config.project_paths import NODEJS_SCRIPT
from vigorish.constants import JOB_SPINNER_COLORS
from vigorish.enums import DataSet, ScrapeCondition
from vigorish.scrape.url_tracker import UrlTracker
from vigorish.util.datetime_util import get_date_range
from vigorish.util.result import Result
from vigorish.util.sys_helpers import execute_nodejs_script


def user_cancelled(db_session, active_job, spinner, signal_received, frame):
    spinner.stop()
    print_message("\nJob cancelled by user!", fg="yellow", bold=True)
    pause(message="Press any key to continue...")
    exit(0)


class ScrapeTaskABC(ABC):
    data_set: DataSet
    url_tracker: UrlTracker

    def __init__(self, app, db_job):
        self.app = app
        self.config = app.config
        self.db_session = app.db_session
        self.scraped_data = app.scraped_data
        self.db_job = db_job
        self.start_date = None
        self.end_date = None
        self.season = self.db_job.season
        self.total_days = self.db_job.total_days
        self.scrape_condition = self.app.get_current_setting("SCRAPE_CONDITION", self.data_set)
        self.spinner = Halo(color=JOB_SPINNER_COLORS[self.data_set], spinner="dots3")

    @property
    def date_range(self):
        return get_date_range(self.start_date, self.end_date)

    def execute(self, start_date=None, end_date=None):
        self.start_date = start_date
        self.end_date = end_date
        if not start_date:
            self.start_date = self.db_job.start_date
        if not end_date:
            self.end_date = self.db_job.end_date

        result = (
            self.initialize()
            .on_success(self.identify_needed_urls)
            .on_success(self.retrieve_scraped_html)
            .on_success(self.scrape_missing_html)
            .on_success(self.parse_scraped_html)
        )
        self.spinner.stop()
        if result.failure:
            return result
        self.url_tracker.remove_scraped_html()
        return Result.Ok()

    def initialize(self):
        if self.scrape_condition == ScrapeCondition.NEVER:
            return Result.Fail("skip")
        signal(SIGINT, partial(user_cancelled, self.db_session, self.db_job, self.spinner))
        self.spinner.text = "Building URL List..."
        self.spinner.start()
        self.url_tracker = UrlTracker(self.db_job, self.data_set, self.scraped_data)
        result = self.url_tracker.create_url_set(self.start_date, self.end_date)
        return result if result.failure else Result.Ok() if self.url_tracker.total_urls else Result.Fail("skip")

    def identify_needed_urls(self):
        self.spinner.text = self.url_tracker.identify_html_report
        for game_date, urls in self.url_tracker.all_urls.items():
            result = self.check_prerequisites(game_date)
            if result.failure:
                return result
            result = self.check_current_status(game_date)
            if result.failure:
                if "skip" not in result.error:
                    return result
                self.url_tracker.skip_urls.extend(urls)
            else:
                self.url_tracker.need_urls.extend(urls)
            self.spinner.text = self.url_tracker.identify_html_report
        return Result.Ok()

    def retrieve_scraped_html(self):
        self.spinner.text = self.url_tracker.retrieve_html_report
        if not self.url_tracker.need_urls:
            return Result.Fail("skip")
        for url in self.url_tracker.need_urls[:]:
            self.scraped_data.get_html(self.data_set, url.url_id)
            if not url.html_was_scraped:
                self.url_tracker.missing_urls.append(url)
            else:
                self.url_tracker.cached_urls.append(url)
            self.url_tracker.need_urls.remove(url)
            self.spinner.text = self.url_tracker.retrieve_html_report
        return Result.Ok()

    def scrape_missing_html(self):
        while not self.url_tracker.html_scraping_complete:
            self.spinner.text = self.url_tracker.scrape_html_report
            self.spinner.stop_and_persist(self.spinner.frame(), "")
            result = self.invoke_nodejs_script()
            self.spinner.text = self.url_tracker.save_html_report
            self.spinner.start()
            for url in self.url_tracker.missing_urls[:]:
                if not url.scraped_html_is_valid:
                    continue
                result = self.scraped_data.save_html(self.data_set, url.url_id, url.html)
                if result.failure:
                    return result
                self.url_tracker.completed_urls.append(url)
                self.url_tracker.missing_urls.remove(url)
                self.spinner.text = self.url_tracker.save_html_report
        return Result.Ok()

    def invoke_nodejs_script(self):
        missing_urls_filepath = self.url_tracker.create_missing_urls_json_file()
        script_args = self.config.get_nodejs_script_args(self.data_set, missing_urls_filepath)
        result = execute_nodejs_script(NODEJS_SCRIPT, script_args)
        missing_urls_filepath.unlink()
        return result

    def parse_scraped_html(self):
        parsed = 0
        self.spinner.text = self.url_tracker.parse_html_report(parsed)
        for urls_for_date in self.url_tracker.all_urls.values():
            for url_details in urls_for_date:
                if url_details.url_id not in self.url_tracker.parse_url_ids:
                    continue
                result = self.parse_html(url_details)
                if result.failure:
                    if "Unable to parse any game data" in result.error:
                        continue
                    return result
                parsed_data = result.value
                result = self.scraped_data.save_json(self.data_set, parsed_data)
                if result.failure:
                    return result
                result = self.update_status(parsed_data)
                if result.failure:
                    return result
                self.db_session.commit()
                parsed += 1
                self.spinner.text = self.url_tracker.parse_html_report(parsed)
        return Result.Ok()

    @abstractmethod
    def check_prerequisites(self, game_date):
        pass

    @abstractmethod
    def check_current_status(self, game_date):
        pass

    @abstractmethod
    def parse_html(self, url_details):
        pass

    @abstractmethod
    def update_status(self, parsed_data):
        pass



import json

from tqdm import tqdm

from vigorish.database import PitchTypePercentile
from vigorish.enums import PitchType
from vigorish.util.result import Result

JSON_FILE_MAP = {
    "thrown_r": "pitch_type_percentiles_rhp.json",
    "thrown_l": "pitch_type_percentiles_lhp.json",
    "thrown_both": "pitch_type_percentiles_both.json",
}


def populate_pitch_type_perc(app, json_folder):
    with tqdm(
        total=len(JSON_FILE_MAP),
        desc="Populating percentiles table...",
        unit="file",
        mininterval=0.12,
        maxinterval=5,
        unit_scale=True,
        ncols=90,
    ) as pbar:
        for pitch_thrown, filename in JSON_FILE_MAP.items():
            json_file = json_folder.joinpath(filename)
            for stat_name, stat_map in json.loads(json_file.read_text()).items():
                for pitch_type_abbrev, pitch_type_map in stat_map.items():
                    pitch_type = PitchType.from_abbrev(pitch_type_abbrev)
                    if PitchType.PERCENTILES & pitch_type == pitch_type:
                        for percentile, stat_value in pitch_type_map.items():
                            percentile_dict = {
                                "stat_name": stat_name,
                                "pitch_type": pitch_type_abbrev,
                                "thrown_r": pitch_thrown == "thrown_r",
                                "thrown_l": pitch_thrown == "thrown_l",
                                "thrown_both": pitch_thrown == "thrown_both",
                                "percentile": percentile,
                                "stat_value": stat_value,
                            }
                            app.db_session.add(PitchTypePercentile(**percentile_dict))
            pbar.update()
    app.db_session.commit()
    return Result.Ok()



"""Individual game info scraped from brooksbaseball.com."""
from dataclasses import dataclass, field
from typing import Dict


@dataclass(eq=True)
class BrooksGameInfo:
    """Individual game info scraped from brooksbaseball.com."""

    might_be_postponed: bool = field(repr=False, default=None)
    game_date_year: str = field(repr=False, default="")
    game_date_month: str = field(repr=False, default="")
    game_date_day: str = field(repr=False, default="")
    game_time_hour: str = field(repr=False, default="")
    game_time_minute: str = field(repr=False, default="")
    time_zone_name: str = field(repr=False, default="")
    bb_game_id: str = field(repr=False, default="")
    bbref_game_id: str = ""
    away_team_id_bb: str = field(repr=False, default="")
    home_team_id_bb: str = field(repr=False, default="")
    game_number_this_day: int = field(repr=False, default=0)
    pitcher_appearance_count: int = field(repr=False, default=0)
    pitcher_appearance_dict: Dict[str, str] = field(repr=False, default_factory=dict)

    @property
    def all_pitch_app_ids(self):
        return [f"{self.bbref_game_id}_{pitch_app_id}" for pitch_app_id in self.pitcher_appearance_dict.keys()]

    def as_dict(self):
        """Convert game info list to a dictionary."""
        return {
            "__brooks_game_info__": True,
            "might_be_postponed": self.might_be_postponed,
            "game_date_year": int(self.game_date_year),
            "game_date_month": int(self.game_date_month),
            "game_date_day": int(self.game_date_day),
            "game_time_hour": int(self.game_time_hour),
            "game_time_minute": int(self.game_time_minute),
            "time_zone_name": self.time_zone_name,
            "bb_game_id": self.bb_game_id,
            "bbref_game_id": self.bbref_game_id,
            "away_team_id_bb": self.away_team_id_bb,
            "home_team_id_bb": self.home_team_id_bb,
            "game_number_this_day": self.game_number_this_day,
            "pitcher_appearance_count": int(self.pitcher_appearance_count),
            "pitcher_appearance_dict": self.pitcher_appearance_dict,
        }



# flake8: noqa
from vigorish.cli.menu_items.change_config_setting import ChangeConfigSettting
from vigorish.cli.menu_items.change_env_var_setting import ChangeEnvVarSetting
from vigorish.cli.menu_items.combine_data import CombineScrapedData
from vigorish.cli.menu_items.config_setting import ConfigSetting
from vigorish.cli.menu_items.create_job import CreateJob
from vigorish.cli.menu_items.exit_program import ExitProgram
from vigorish.cli.menu_items.investigate_data_failures import InvestigateScrapedDataFailures
from vigorish.cli.menu_items.investigate_invalid_pfx import InvestigateInvalidPitchFx
from vigorish.cli.menu_items.investigate_pitchfx_errors import InvestigatePitchFxErrors
from vigorish.cli.menu_items.job_details import JobDetails
from vigorish.cli.menu_items.return_to_parent import ReturnToParent
from vigorish.cli.menu_items.setup_db import SetupDatabase
from vigorish.cli.menu_items.status_report import StatusReport
from vigorish.cli.menu_items.view_game_data import ViewGameData
